"""Wherein is contained the concrete SQLite implementation of the Index."""
# Pylint doesn't like the similarity between this file and the SQLite file, but
# it doesn't make sense to write shared functions for them. So ignore pylint.
# pylint: disable=duplicate-code
import json
import os
import warnings
from datetime import datetime

import ast
import dateutil
import pandas as pd
# Try-Except required to make pyodbc/sqlalchemy an optional dependency.
try:
    # pylint: disable=unused-import
    # pyodbc is imported here because sqlalchemy requires it.
    import pyodbc # noqa: F401
    import sqlalchemy as sqla # noqa: F401
    # pylint: enable=unused-import
except ImportError:
    _HAS_REQUIRED_DEPS = False
else:
    _HAS_REQUIRED_DEPS = True

from .index_abc import IndexABC
from .list_baskets import _get_list_of_basket_jsons
from .validate_basket import validate_basket_dict

class IndexSQL(IndexABC):
    """Concrete implementation of Index, using SQL."""
    def __init__(self, file_system, pantry_path, **kwargs):
        """Initializes the Index class.

        Parameters
        ----------
        file_system: fsspec object
            The fsspec object which hosts the pantry we desire to index.
        pantry_path: str
            Path to the pantry root which we want to index.
        **database_name: str (default='weave_db')
            DB to be used. If none is set, defaults to 'weave_db'.
        **encrypt: bool (default=False)
            Whether or not to encrypt the database (disabled by default, due
            to the fact that it requires additional configuration).
        **odbc_driver: str (default='ODBC Driver 18 for SQL Server')
            Driver to be used for the connection. Usually similar to:
            'ODBC Driver <18, 17, 13...> for SQL Server'.
        **pantry_schema: str (default=<pantry_path>)
            The schema to use for the pantry. If none is set, defaults to the
            pantry path (with _ replacements when necessary).
        """
        if not _HAS_REQUIRED_DEPS:
            raise ImportError("Missing Dependencies. The packages: "
                              "'sqlalchemy' are required to use this class")

        # Check that the required environment variables are set.
        try:
            # We want to fail early if these are not set, so we check them here
            # even though they are not used until we connect.
            # Pylint thinks this is pointless, so pylint is getting ignored.
            # pylint: disable=pointless-statement
            os.environ["MSSQL_HOST"]
            os.environ["MSSQL_USERNAME"]
            os.environ["MSSQL_PASSWORD"]
            # pylint: enable=pointless-statement
        except KeyError as key_error:
            raise KeyError("The following environment variables must be set "
                           "to use this class: MSSQL_HOST, MSSQL_USERNAME, "
                           "MSSQL_PASSWORD.") from key_error

        self._file_system = file_system
        self._pantry_path = pantry_path

        # Set the database name (defaults to weave_db). DATABASE MUST ALREADY
        # EXIST. If it does not, the user must create it manually.
        self._database_name = kwargs.get("database_name", "weave_db")\
            .replace("-", "_")

        # Set the schema name (defaults to pantry_path). If the schema does not
        # exist, it will be created.
        d_schema_name = self._pantry_path.replace(os.sep, "_")\
            .replace("-", "_")
        if d_schema_name == "":
            d_schema_name = "weave"
        self._pantry_schema = kwargs.get("pantry_schema", d_schema_name)

        self._engine = sqla.create_engine(sqla.engine.url.URL(
            drivername="mssql+pyodbc",
            username=os.environ["MSSQL_USERNAME"],
            password=os.environ["MSSQL_PASSWORD"],
            host=os.environ["MSSQL_HOST"],
            database=self.database_name,
            query={
                "driver": kwargs.get("odbc_driver",
                                     "ODBC Driver 18 for SQL Server"),
                "Encrypt": "yes" if kwargs.get("encrypt", False) else "no",
            },
            port=os.environ.get("MSSQL_PORT", "1433"),
        ))

        self._create_schema()
        self._create_tables()

    # Create read-only properties for the database and schema names.
    @property
    def database_name(self):
        """The database name of the index."""
        return self._database_name

    @property
    def pantry_schema(self):
        """The schema name of the index."""
        return self._pantry_schema

    def execute_sql(self, sql_query, params=None, commit=False):
        """Executes the given SQL query. Returns the results.

        Parameters
        ----------
        sql_query: str or sqlalchemy.sql.text
            The SQL query to be executed.
        params: dict (optional)
            The parameters to be used in the query.
        commit: bool (default=False)
            Whether or not to commit the query.

        Returns
        ----------
        (list, list) or int or None
            Returns the results of the query and the column names as a tuple.
            If statement affects rows, returns the number of rows affected.
            If the query does not return any results, returns None.
        """
        try:
            with self._engine.connect() as connection:
                if isinstance(sql_query, sqla.sql.elements.TextClause):
                    query = sql_query
                elif isinstance(sql_query, str):
                    query = sqla.sql.text(sql_query)
                else:
                    raise ValueError(
                        "sql_query should be a str or a "
                        "sqlalchemy TextClause object"
                    )

                if params is not None:
                    if not isinstance(params, dict):
                        raise TypeError("params should be a dict")

                    # Execute the SQL query with parameters
                    result = connection.execute(query, params)
                else:
                    # Execute the SQL query without parameters
                    result = connection.execute(query)

                if commit:
                    connection.commit()

                # Fetch and return the results
                if result.returns_rows:
                    return result.fetchall(), list(result.keys())

                # Return rows affected (used for INSERT, DELETE, etc.)
                if result.rowcount != -1:
                    return result.rowcount

                return None

        except sqla.exc.SQLAlchemyError as err:
            raise err

    def _create_schema(self):
        """Create the schema if it does not already exist."""
        # Check if self.pantry_schema exists.
        results, _ = self.execute_sql(
            "SELECT * FROM sys.schemas WHERE name = :pantry_schema;",
            {"pantry_schema": self.pantry_schema}
        )
        # If it does not exist, create it.
        if not results:
            self.execute_sql(f"CREATE SCHEMA {self.pantry_schema};",
                             commit=True)

    def _create_tables(self):
        """Create the required tables if they do not already exist."""
        # THIS NEEDS TO BE UPDATED MANUALLY IF NEW COLUMNS ARE ADDED TO INDEX.
        # THE INSERT IN OTHER FUNCTIONS USE config.index_schema(), BUT THAT
        # CAN'T BE USED HERE AS TYPE NEEDS TO BE SPECIFIED.
        self.execute_sql(
            f"""
            IF NOT EXISTS (
                SELECT * FROM sys.tables t
                JOIN sys.schemas s ON (t.schema_id = s.schema_id)
                WHERE s.name = '{self.pantry_schema}'
                AND t.name = 'pantry_index'
            )
            CREATE TABLE {self.pantry_schema}.pantry_index (
                uuid varchar(64),
                upload_time INT,
                parent_uuids TEXT,
                basket_type TEXT,
                label TEXT,
                weave_version TEXT,
                address TEXT,
                storage_type TEXT,
                PRIMARY KEY(uuid),
                UNIQUE(uuid)
            );
            """, commit=True
        )

        self.execute_sql(
            f"""
            IF NOT EXISTS (
                SELECT * FROM sys.tables t
                JOIN sys.schemas s ON (t.schema_id = s.schema_id)
                WHERE s.name = '{self.pantry_schema}'
                AND t.name = 'parent_uuids'
            )
            CREATE TABLE {self.pantry_schema}.parent_uuids (
                uuid varchar(64),
                parent_uuid varchar(64),
                PRIMARY KEY(uuid, parent_uuid),
                UNIQUE(uuid, parent_uuid)
            );
            """, commit=True
        )

    @property
    def file_system(self):
        """The file system of the pantry referenced by this Index."""
        return self._file_system

    @property
    def pantry_path(self):
        """The pantry path referenced by this Index."""
        return self._pantry_path

    def generate_metadata(self, **kwargs):
        """Populates the metadata for the index.

        Parameters
        ----------
        **kwargs unused for this function.

        Returns
        ----------
        dict
            Returns a dictionary of the metadata.
        """
        return {
            "database_host": os.environ["MSSQL_HOST"],
            "database_port": os.environ.get("MSSQL_PORT", "1433"),
            "database_name": self.database_name,
            "database_schema": self.pantry_schema,
        }

    def generate_index(self, **kwargs):
        """Populates the index from the file system.

        Generate the index by scraping the pantry and adding the manifest data
        of found baskets to the index.

        Parameters
        ----------
        **kwargs unused for this function.
        """
        if not isinstance(self.pantry_path, str):
            raise TypeError("'pantry_path' must be a string: "
                            f"'{self.pantry_path}'")

        if not self.file_system.exists(self.pantry_path):
            raise FileNotFoundError("'pantry_path' does not exist: "
                                    f"'{self.pantry_path}'")

        basket_jsons = _get_list_of_basket_jsons(
            self.pantry_path,
            self.file_system,
        )

        bad_baskets = []
        for basket_json_address in basket_jsons:
            with self.file_system.open(basket_json_address, "rb") as file:
                basket_dict = json.load(file)

                # If the basket is invalid, add it to a list, then skip it.
                if not validate_basket_dict(basket_dict):
                    bad_baskets.append(os.path.dirname(basket_json_address))
                    continue
                # Skip baskets that are indexes.
                if basket_dict["basket_type"] == "index":
                    continue

                absolute_path = os.path.dirname(basket_json_address)
                relative_path = (
                    absolute_path[absolute_path.find(self.pantry_path):]
                )
                basket_dict["address"] = relative_path
                basket_dict["storage_type"] = (
                    self.file_system.__class__.__name__
                )
                basket_dict["upload_time"] = int(
                    datetime.timestamp(
                        dateutil.parser.parse(basket_dict["upload_time"])
                    )
                )

                parent_uuids = basket_dict["parent_uuids"]
                basket_dict["parent_uuids"] = str(basket_dict["parent_uuids"])

                if "weave_version" not in basket_dict.keys():
                    basket_dict["weave_version"] = "<0.13.0"

                # Insert into pantry_index.
                index_columns = list(basket_dict.keys())
                sql = sqla.text(
                    f"INSERT INTO {self.pantry_schema}.pantry_index ("
                    f"{', '.join([f'{column}' for column in index_columns])}) "
                    "SELECT "
                    f"{', '.join([f':{column}' for column in index_columns])} "
                    "WHERE NOT EXISTS "
                    f"(SELECT 1 FROM {self.pantry_schema}.pantry_index "
                    "WHERE uuid = :uuid);"
                )
                self.execute_sql(sql, basket_dict, commit=True)

                # Insert into parent_uuids.
                sql = sqla.text(
                    f"INSERT INTO {self.pantry_schema}.parent_uuids "
                    "(uuid, parent_uuid) "
                    "SELECT :uuid, :parent_uuid "
                    "WHERE NOT EXISTS "
                    f"(SELECT 1 FROM {self.pantry_schema}.parent_uuids "
                    "WHERE uuid = :uuid AND parent_uuid = :parent_uuid);"
                )
                # Loop all uuids and parent uuids.
                for parent_uuid in parent_uuids:
                    self.execute_sql(
                        sql,
                        {
                            "uuid": basket_dict["uuid"],
                            "parent_uuid": parent_uuid,
                        },
                        commit=True,
                    )

        if len(bad_baskets) != 0:
            warnings.warn("baskets found in the following locations "
                          "do not follow specified weave schema:\n"
                          f"{bad_baskets}")

    def to_pandas_df(self, max_rows=1000, **kwargs):
        """Returns the pandas dataframe representation of the index.

        Parameters
        ----------
        max_rows: int (default=1000)
            Max rows returned in the pandas dataframe.
        **kwargs unused for this function.

        Returns
        ----------
        pandas.DataFrame
            Returns a dataframe of the manifest data of the baskets in the
            pantry.
        """
        # Get the rows from the index as a list of lists, then get the columns.
        result, columns = self.execute_sql(
            f"SELECT TOP (:max_rows) * FROM {self.pantry_schema}.pantry_index",
            {"max_rows": max_rows},
        )
        result = [list(row) for row in result]

        ind_df = pd.DataFrame(result, columns=columns)
        ind_df["parent_uuids"] = ind_df["parent_uuids"].apply(ast.literal_eval)
        ind_df["upload_time"] = pd.to_datetime(
            ind_df["upload_time"],
            unit="s",
            origin="unix",
        )
        return ind_df

    def track_basket(self, entry_df, **kwargs):
        """Track a basket (or many baskets) from the pantry with the Index.

        Parameters
        ----------
        entry_df: pd.DataFrame
            Uploaded baskets' manifest data to append to the index.

        **kwargs unused for this function.
        """
        # Save off the original uuid and parent_uuids columns.
        uuids = entry_df["uuid"]
        parent_uuids = entry_df["parent_uuids"]

        # Convert the parent_uuids to a string, and the upload_time to an int.
        entry_df["parent_uuids"] = entry_df["parent_uuids"].astype(str)
        entry_df["upload_time"] = (
            entry_df["upload_time"].astype(int) // 1e9
        ).astype(int)

        for basket_dict in entry_df.to_dict(orient="records"):
            index_columns = list(basket_dict.keys())
            # Insert into pantry_index.
            sql = sqla.text(
                f"INSERT INTO {self.pantry_schema}.pantry_index ("
                f"{', '.join([f'{column}' for column in index_columns])}) "
                "SELECT "
                f"{', '.join([f':{column}' for column in index_columns])} "
                "WHERE NOT EXISTS "
                f"(SELECT 1 FROM {self.pantry_schema}.pantry_index "
                "WHERE uuid = :uuid);"
            )
            self.execute_sql(sql, basket_dict, commit=True)

            # Insert into parent_uuids.
            sql = sqla.text(
                f"INSERT INTO {self.pantry_schema}.parent_uuids "
                "(uuid, parent_uuid) "
                "SELECT :uuid, :parent_uuid "
                "WHERE NOT EXISTS "
                f"(SELECT 1 FROM {self.pantry_schema}.parent_uuids "
                "WHERE uuid = :uuid AND parent_uuid = :parent_uuid);"
            )
            # Loop all uuids and parent uuids (list of lists).
            for uuid, parent_uuids in zip(uuids, parent_uuids):
                # Loop all parent uuids (now a list of strings)
                for parent_uuid in parent_uuids:
                    self.execute_sql(
                        sql,
                        {
                            "uuid": uuid,
                            "parent_uuid": parent_uuid,
                        },
                        commit=True,
                    )

    def untrack_basket(self, basket_address, **kwargs):
        """Remove a basket from being tracked of given UUID or path.

        Parameters
        ----------
        basket_address: str or [str]
            Argument can take one of two forms: either a path to the basket
            directory, or the UUID of the basket. These may also be passed in
            as a list.

        **kwargs unused for this function.
        """
        if not isinstance(basket_address, list):
            basket_address = [basket_address]

        if self.file_system.exists(os.fspath(basket_address[0])):
            uuids, _ = self.execute_sql(sqla.text(
                f"SELECT uuid FROM {self.pantry_schema}.pantry_index "
                "WHERE CONVERT(nvarchar(MAX), address) IN "
                "(SELECT CONVERT(nvarchar(MAX), value) "
                "FROM STRING_SPLIT(:basket_address, ','))"),
                {"basket_address": ','.join(basket_address)}
            )
            uuids = [uuid[0] for uuid in uuids]
        else:
            uuids = basket_address

        # Delete from pantry_index.
        query = (
            f"DELETE FROM {self.pantry_schema}.pantry_index WHERE "
            "CONVERT(nvarchar(MAX), uuid) IN "
            "(SELECT CONVERT(nvarchar(MAX), value) "
            "FROM STRING_SPLIT(:uuids, ','))"
        )
        rowcount = self.execute_sql(
            query,
            {"uuids": ','.join(uuids)},
            commit=True,
        )
        if rowcount != len(uuids):
            warnings.warn(
                UserWarning(
                    "Incomplete Request. Index could not untrack baskets, "
                    "as some were not being tracked to begin with.",
                    len(uuids) - rowcount
                )
            )

        # Delete from parent_uuids.
        query = (
            f"DELETE FROM {self.pantry_schema}.parent_uuids WHERE "
            "CONVERT(nvarchar(MAX), uuid) IN "
            "(SELECT CONVERT(nvarchar(MAX), value) "
            "FROM STRING_SPLIT(:uuids, ','))"
        )
        self.execute_sql(query, {"uuids": ','.join(uuids)}, commit=True)

    def get_rows(self, basket_address, **kwargs):
        """Returns a pd.DataFrame row information of given UUID or path.

        Parameters
        ----------
        basket_address: str or [str]
            Argument can take one of two forms: either a path to the basket
            directory, or the UUID of the basket. These may also be passed in
            as a list.
        **kwargs unused for this function.

        Returns
        ----------
        pandas.DataFrame
            Manifest information for the requested basket(s).
        """
        if not isinstance(basket_address, list):
            basket_address = [basket_address]

        if self.file_system.exists(os.fspath(basket_address[0])):
            id_column = "address"
        else:
            id_column = "uuid"

        query = (
            f"SELECT * FROM {self.pantry_schema}.pantry_index "
            f"WHERE CONVERT(nvarchar(MAX), {id_column}) IN "
            "(SELECT CONVERT(nvarchar(MAX), value) "
            "FROM STRING_SPLIT(:basket_address, ','))"
        )
        results, columns = self.execute_sql(
            query,
            {"basket_address": ','.join(basket_address)},
        )
        results = [list(row) for row in results]

        ind_df = pd.DataFrame(
            results,
            columns=columns
        )
        ind_df["parent_uuids"] = ind_df["parent_uuids"].apply(ast.literal_eval)
        ind_df["upload_time"] = pd.to_datetime(
            ind_df["upload_time"],
            unit="s",
            origin="unix",
        )
        return ind_df

    def get_parents(self, basket_address, **kwargs):
        """Returns a pandas dataframe of all parents of a basket.

        Parameters
        ----------
        basket_address: str
            Argument can take one of two forms: either a path to the basket
            directory, or the UUID of the basket.
        **max_gen_level: int (optional)
            This indicates the maximum generation level that will be reported.

        Returns
        ----------
        pandas.DataFrame containing all the manifest data AND generation level
        of parents (and recursively their parents) of the given basket.
        """

        max_gen_level = kwargs.get("max_gen_level", 999)

        if self.file_system.exists(os.fspath(basket_address)):
            id_column = "address"
        else:
            id_column = "uuid"

        basket_uuid, _ = self.execute_sql(
            f"SELECT uuid FROM {self.pantry_schema}.pantry_index "
            f"WHERE CONVERT(nvarchar(MAX), {id_column}) = :basket_address",
            {"basket_address": basket_address}
        )

        if basket_uuid is None or len(basket_uuid) == 0:
            raise FileNotFoundError(
                f"basket path or uuid does not exist '{basket_address}'"
            )
        basket_uuid = basket_uuid[0][0]

        results, columns = self.execute_sql(
            f"""
            WITH child_record AS (
                SELECT
                    0 AS level,
                    CAST(:basket_uuid AS nvarchar(max)) AS id,
                    CAST(:basket_uuid AS nvarchar(max)) AS path
                UNION ALL
                SELECT
                    child_record.level + 1,
                    CAST(parent_uuids.parent_uuid AS nvarchar(max)) AS id,
                    CAST(child_record.path + '/' + parent_uuids.parent_uuid
                        AS nvarchar(max)) AS path
                FROM {self.pantry_schema}.parent_uuids
                JOIN child_record ON parent_uuids.uuid = child_record.id
                WHERE 
                    path NOT LIKE CONCAT(parent_uuids.parent_uuid, '/%')
                    AND path NOT LIKE CONCAT('%', parent_uuids.parent_uuid)
                    AND path NOT LIKE
                        CONCAT('%', parent_uuids.parent_uuid, '/%')
                AND child_record.level < :max_gen_level
            )
            SELECT pantry_index.*, child_record.level, child_record.path
            FROM {self.pantry_schema}.pantry_index as pantry_index
            JOIN child_record ON pantry_index.uuid = child_record.id
            ORDER BY child_record.level ASC OPTION (MAXRECURSION 0);
            """,
            {"basket_uuid": basket_uuid, "max_gen_level": max_gen_level}
        )

        results = [list(row) for row in results]
        columns[columns.index("level")] = "generation_level"

        parent_df = pd.DataFrame(results, columns=columns)
        parent_df = parent_df.drop_duplicates()
        parent_df = parent_df[parent_df["uuid"] != basket_uuid]

        if parent_df.empty:
            return parent_df
        parent_df["parent_uuids"] = parent_df["parent_uuids"].apply(
            ast.literal_eval
        )
        parent_df["upload_time"] = pd.to_datetime(
            parent_df["upload_time"],
            unit="s",
            origin="unix",
        )

        for _, row in parent_df.iterrows():
            for prev in row['path'].split('/'):
                if prev in row["parent_uuids"]:
                    raise ValueError(
                        f"Parent-Child loop found at uuid: {basket_uuid}"
                    )

        parent_df.drop(columns="path", inplace=True)

        return parent_df

    def get_children(self, basket_address, **kwargs):
        """Returns a pandas dataframe of all children of a basket.

        Parameters
        ----------
        basket_address: str
            Argument can take one of two forms: either a path to the basket
            directory, or the UUID of the basket.
        **min_gen_level: int (optional)
            This indicates the minimum generation level that will be reported.

        Returns
        ----------
        pandas.DataFrame containing all the manifest data AND generation level
        of children (and recursively their children) of the given basket.
        """

        min_gen_level = kwargs.get("min_gen_level", -999)

        if self.file_system.exists(os.fspath(basket_address)):
            id_column = "address"
        else:
            id_column = "uuid"

        basket_uuid, _ = self.execute_sql(
            f"SELECT uuid FROM {self.pantry_schema}.pantry_index "
            f"WHERE CONVERT(nvarchar(MAX), {id_column}) = :basket_address",
            {"basket_address": basket_address}
        )

        if basket_uuid is None or len(basket_uuid) == 0:
            raise FileNotFoundError(
                f"basket path or uuid does not exist '{basket_address}'"
            )
        basket_uuid = basket_uuid[0][0]

        results, columns = self.execute_sql(
            f"""
            WITH child_record AS (
                SELECT
                    0 AS level,
                    CAST(:basket_uuid AS nvarchar(max)) AS id,
                    CAST(:basket_uuid AS nvarchar(max)) AS path
                UNION ALL
                SELECT 
                    child_record.level - 1,
                    CAST(parent_uuids.uuid AS nvarchar(max)) AS id,
                    CAST(child_record.path + '/' + parent_uuids.uuid
                        AS nvarchar(max)) AS path
                FROM {self.pantry_schema}.parent_uuids
                JOIN child_record ON parent_uuids.parent_uuid = child_record.id
                WHERE 
                    path NOT LIKE CONCAT(parent_uuids.uuid, '/%')
                    AND path NOT LIKE CONCAT('%', parent_uuids.uuid)
                    AND path NOT LIKE CONCAT('%', parent_uuids.uuid, '/%')
                AND child_record.level > :min_gen_level
            )
            SELECT pantry_index.*, child_record.level, child_record.path
            FROM {self.pantry_schema}.pantry_index as pantry_index
            JOIN child_record ON pantry_index.uuid = child_record.id
            ORDER BY child_record.level DESC OPTION (MAXRECURSION 0);
            """,
            {"basket_uuid": basket_uuid, "min_gen_level": min_gen_level}
        )

        results = [list(row) for row in results]
        columns[columns.index("level")] = "generation_level"

        child_df = pd.DataFrame(results, columns=columns)
        child_df = child_df.drop_duplicates()
        child_df["parent_uuids"] = child_df["parent_uuids"].apply(
            ast.literal_eval
        )
        child_df["upload_time"] = pd.to_datetime(
            child_df["upload_time"],
            unit="s",
            origin="unix",
        )

        parents = {}
        for _, row in child_df.iterrows():
            parents[row['uuid']] = row["parent_uuids"]
            for prev in row['path'].split('/'):
                if row['uuid'] in parents[prev]:
                    raise ValueError(
                        f"Parent-Child loop found at uuid: {basket_uuid}"
                    )

        child_df = child_df[child_df['uuid'] != basket_uuid]
        child_df.drop(columns="path", inplace=True)

        return child_df

    def get_baskets_of_type(self, basket_type, max_rows=1000, **kwargs):
        """Returns a pandas dataframe containing baskets of basket_type.

        Parameters
        ----------
        basket_type: str
            The basket type to filter for.
        max_rows: int (default=1000)
            Max rows returned in the pandas dataframe.

        **kwargs unused for this function.

        Returns
        ----------
        pandas.DataFrame containing the manifest data of baskets of the type.
        """
        result, columns = self.execute_sql(
            f"SELECT TOP (:max_rows) * FROM {self.pantry_schema}.pantry_index "
            "WHERE CONVERT(nvarchar(MAX), basket_type) = :basket_type",
            {"max_rows": max_rows, "basket_type": basket_type},
        )
        result = [list(row) for row in result]

        ind_df = pd.DataFrame(result, columns=columns)
        ind_df["parent_uuids"] = ind_df["parent_uuids"].apply(ast.literal_eval)
        ind_df["upload_time"] = pd.to_datetime(
            ind_df["upload_time"],
            unit="s",
            origin="unix",
        )
        return ind_df

    def get_baskets_of_label(self, basket_label, max_rows=1000, **kwargs):
        """Returns a pandas dataframe containing baskets with label.

        Parameters
        ----------
        basket_label: str
            The label to filter for.
        max_rows: int (default=1000)
            Max rows returned in the pandas dataframe.

        **kwargs unused for this function.

        Returns
        ----------
        pandas.DataFrame containing the manifest data of baskets with the label
        """
        result, columns = self.execute_sql(
            f"SELECT TOP (:max_rows) * FROM {self.pantry_schema}.pantry_index "
            "WHERE CONVERT(nvarchar(MAX), label) = :basket_label",
            {"max_rows": max_rows, "basket_label": basket_label},
        )
        result = [list(row) for row in result]

        ind_df = pd.DataFrame(result, columns=columns)
        ind_df["parent_uuids"] = ind_df["parent_uuids"].apply(ast.literal_eval)
        ind_df["upload_time"] = pd.to_datetime(
            ind_df["upload_time"],
            unit="s",
            origin="unix",
        )
        return ind_df

    def get_baskets_by_upload_time(self, start_time=None, end_time=None,
                                   max_rows=1000, **kwargs):
        """Returns a pandas dataframe of baskets uploaded between two times.

        Parameters
        ----------
        start_time: datetime.datetime (optional)
            The start datetime object to filter between. If None, will filter
            from the beginning of time.
        end_time: datetime.datetime (optional)
            The end datetime object to filter between. If None, will filter
            to the current datetime.
        max_rows: int (default=1000)
            Max rows returned in the pandas dataframe.

        **kwargs unused for this function.

        Returns
        ----------
        pandas.DataFrame containing the manifest data of baskets uploaded
        between the start and end times.
        """
        super().get_baskets_by_upload_time(start_time, end_time)
        if start_time is None and end_time is None:
            return self.to_pandas_df(max_rows=max_rows)

        if start_time and end_time:
            start_time = int(datetime.timestamp(start_time))
            end_time = int(datetime.timestamp(end_time))
            results, columns = self.execute_sql(
                f"""SELECT TOP (:max_rows) *
                FROM {self.pantry_schema}.pantry_index
                WHERE upload_time >= :start_time AND upload_time <= :end_time
                """,
                {"max_rows": max_rows,
                 "start_time": start_time,
                 "end_time": end_time
                })
        elif start_time:
            start_time = int(datetime.timestamp(start_time))
            results, columns = self.execute_sql(
                f"""SELECT TOP (:max_rows) *
                FROM {self.pantry_schema}.pantry_index
                WHERE upload_time >= :start_time
                """, {"max_rows": max_rows, "start_time": start_time})
        elif end_time:
            end_time = int(datetime.timestamp(end_time))
            results, columns = self.execute_sql(
                f"""SELECT TOP (:max_rows) *
                FROM {self.pantry_schema}.pantry_index
                WHERE upload_time <= :end_time
                """, {"max_rows": max_rows, "end_time": end_time})

        results = [list(row) for row in results]

        ind_df = pd.DataFrame(results, columns=columns)
        ind_df["parent_uuids"] = ind_df["parent_uuids"].apply(ast.literal_eval)
        ind_df["upload_time"] = pd.to_datetime(
            ind_df["upload_time"],
            unit="s",
            origin="unix",
        )
        return ind_df

    def query(self, expr, **kwargs):
        """Returns a pandas dataframe of the results of the expression.

        Parameters
        ----------
        expr: str
            An expression passed to the backend. An example could be a SQL or
            pandas query. Largely dependent on concrete implementations.
        **expr_args: dict (optional)
            Arguments to pass to the SQL expression. Should be in the form:
            {"arg_name": arg_value, ...}

        Returns
        ----------
        pandas.DataFrame of the resulting query.
        """
        expr_args = kwargs.get("expr_args", ())
        results, columns = self.execute_sql(expr, expr_args)
        results = [list(row) for row in results]
        return pd.DataFrame(results, columns=columns)

    def __len__(self):
        """Returns the number of baskets in the index."""
        results, _ = self.execute_sql(
                f"SELECT COUNT (*) FROM {self.pantry_schema}.pantry_index"
            )
        return results[0][0]

    def __str__(self):
        """Returns the str instantiation type of this Index (ie 'IndexSQL')."""
        return self.__class__.__name__



"""Pytests for the uploader functionality."""

import json
import os
import time
import uuid
from datetime import datetime
from unittest.mock import patch

import pytest
import s3fs
from fsspec.implementations.local import LocalFileSystem

import weave
from weave import Pantry, IndexPandas, IndexSQLite
from weave.tests.pytest_resources import PantryForTest, file_path_in_list
from weave.upload import (
    UploadBasket,
    derive_integrity_data,
    validate_upload_item,
)

# This module is long and has many tests. Pylint is complaining that it is too
# long. This isn't necessarily bad in this case, as the alternative
# would be to write the tests continuuing in a different script, which would
# be unnecesarily complex.
# Disabling this warning for this script.
# pylint: disable=too-many-lines


class UploadForTest(PantryForTest):
    """Test class extended from PantryForTest to include custom call for
    upload.
    """

    def __init__(self, tmpdir, file_system):
        super().__init__(tmpdir, file_system)
        self.uploaded_files = None

    def run_uploader(self, tmp_basket_dir):
        """Wrapper to call the weave upload function.
        """

        upload_items = [
            {
                "path": str(os.path.join(tmp_basket_dir, "test.txt")),
                "stub": False,
            }
        ]
        basket_type = "test-basket"
        metadata = {"oh": "I don't know", "something": "stupid"}
        label = "my label"
        parent_ids = [uuid.uuid1().hex]

        upload_path = weave.upload.UploadBasket(
            upload_items=upload_items,
            basket_type=basket_type,
            pantry_path=self.pantry_path,
            parent_ids=parent_ids,
            metadata=metadata,
            label=label,
            file_system=self.file_system,
        ).get_upload_path()

        self.uploaded_files = self.file_system.ls(upload_path)

        return upload_path


s3fs = s3fs.S3FileSystem(
    client_kwargs={"endpoint_url": os.environ["S3_ENDPOINT"]}
)
local_fs = LocalFileSystem()


# Test with two different fsspec file systems (above).
@pytest.fixture(
    params=[s3fs, local_fs],
    ids=["S3FileSystem", "LocalFileSystem"],
)
def set_up_tu(request, tmpdir):
    """Sets up the test uploader."""
    file_system = request.param
    test_upload = UploadForTest(tmpdir, file_system)
    yield test_upload
    test_upload.cleanup_pantry()


# Ignoring pylint's warning "redefined-outer-name" as this is simply
# how pytest works when it comes to pytest fixtures.
# pylint: disable=redefined-outer-name


def test_upload_test_txt_in_uploaded_files(set_up_tu):
    """Test that uploaded test files are properly uploaded."""
    test_uploader = set_up_tu

    tmp_basket_dir_name = "test_basket_tmp_dir"
    tmp_basket_dir = test_uploader.set_up_basket(tmp_basket_dir_name)
    upload_path = test_uploader.run_uploader(tmp_basket_dir)

    assert file_path_in_list(
        os.path.join(upload_path, "test.txt"), test_uploader.uploaded_files
    )


def test_upload_basket_manifest_in_uploaded_files(set_up_tu):
    """Test that basket manifest files are properly uploaded."""
    test_uploader = set_up_tu

    tmp_basket_dir_name = "test_basket_tmp_dir"
    tmp_basket_dir = test_uploader.set_up_basket(tmp_basket_dir_name)
    upload_path = test_uploader.run_uploader(tmp_basket_dir)

    assert file_path_in_list(
        os.path.join(upload_path, "basket_manifest.json"),
        test_uploader.uploaded_files,
    )


def test_upload_basket_supplement_in_uploaded_files(set_up_tu):
    """Test that basket supplement files are properly uploaded."""
    test_uploader = set_up_tu

    tmp_basket_dir_name = "test_basket_tmp_dir"
    tmp_basket_dir = test_uploader.set_up_basket(tmp_basket_dir_name)
    upload_path = test_uploader.run_uploader(tmp_basket_dir)

    assert file_path_in_list(
        os.path.join(upload_path, "basket_supplement.json"),
        test_uploader.uploaded_files,
    )


def test_upload_basket_metadata_in_uploaded_files(set_up_tu):
    """Test that basket metadata files are properly uploaded."""
    test_uploader = set_up_tu

    tmp_basket_dir_name = "test_basket_tmp_dir"
    tmp_basket_dir = test_uploader.set_up_basket(tmp_basket_dir_name)
    upload_path = test_uploader.run_uploader(tmp_basket_dir)

    assert file_path_in_list(
        os.path.join(upload_path, "basket_metadata.json"),
        test_uploader.uploaded_files,
    )


def test_upload_nothing_else_in_uploaded_files(set_up_tu):
    """Test that only basket data and required files are uploaded."""
    test_uploader = set_up_tu

    tmp_basket_dir_name = "test_basket_tmp_dir"
    tmp_basket_dir = test_uploader.set_up_basket(tmp_basket_dir_name)
    test_uploader.run_uploader(tmp_basket_dir)

    assert len(test_uploader.uploaded_files) == 4


def test_upload_pantry_path_is_string():
    """Test that an error is raised when the pantry name is not a string.
    """

    pantry_path = 7
    upload_items = [
        {"path": "this/doesnt/actually/matter/here", "stub": False}
    ]

    with pytest.raises(
        TypeError,
        match="Invalid datatype: 'pantry_path: must be type <class 'str'>'",
    ):
        weave.upload.UploadBasket(
            upload_items, basket_type="test_basket", pantry_path=pantry_path
        )


def test_validate_upload_item_correct_schema_path_key():
    """Test that validate_upload_item raises a KeyError when an invalid
    path key is used.
    """

    file_path = "path/path"

    # Invalid Path Key
    upload_item = {"path_invalid_key": file_path, "stub": True}
    with pytest.raises(
        KeyError, match="Invalid upload_item key: 'path_invalid_key'"
    ):
        validate_upload_item(upload_item)


def test_validate_upload_item_correct_schema_path_type():
    """Test that validate_upload_item raises a KeyError when an invalid upload
    item type is used.
    """

    upload_item = {"path": 1234, "stub": True}
    with pytest.raises(
        TypeError, match="Invalid upload_item type: 'path: <class 'int'>'"
    ):
        validate_upload_item(upload_item)


def test_validate_upload_item_correct_schema_stub_key():
    """Test that validate_upload_item raises a KeyError when an invalid stub
    key is used.
    """

    file_path = "path/path"
    # Invalid Stub Key
    upload_item = {"path": file_path, "invalid_stub_key": True}
    with pytest.raises(
        KeyError, match="Invalid upload_item key: 'invalid_stub_key'"
    ):
        validate_upload_item(upload_item)


def test_validate_upload_item_correct_schema_stub_type():
    """Test that validate_upload_item raises a KeyError when an invalid stub
    value type is used.
    """

    # Invalid Stub Type
    file_path = "path/path"
    upload_item = {"path": file_path, "stub": "invalid type"}
    with pytest.raises(
        TypeError, match="Invalid upload_item type: 'stub: <class 'str'>'"
    ):
        validate_upload_item(upload_item)


def test_validate_upload_item_correct_schema_extra_key():
    """Test that validate_upload_item raises a KeyError when an invalid
    extra key is used.
    """

    file_path = "path/path"
    # Extra Key
    upload_item = {"path": file_path, "stub": True, "extra_key": True}
    with pytest.raises(KeyError, match="Invalid upload_item key: 'extra_key'"):
        validate_upload_item(upload_item)


def test_validate_upload_item_valid_inputs(tmp_path):
    """Test that no errors are raised when calling validate_upload_item on
    valid inputs.
    """

    text_file_name = "test.txt"
    text_file_content = "0123456789"

    test_file = tmp_path / text_file_name
    test_file.write_text(text_file_content)

    valid_upload_item = {"path": str(test_file), "stub": True}

    validate_upload_item(valid_upload_item)


def test_validate_upload_item_file_exists():
    """Test that validate_upload_item raises a FileExistsError when an invalid
    path value is used.
    """

    upload_item = {"path": "i n v a l i d p a t h", "stub": True}
    with pytest.raises(
        FileExistsError,
        match="'path' does not exist: 'i n v a l i d p a t h'",
    ):
        validate_upload_item(upload_item)


def test_validate_upload_item_folder_exists(tmp_path):
    """Test that validate_upload_item does not raise an error when using a
    folder path.
    """

    text_file_name = "test.txt"
    text_file_content = "0123456789"

    test_file = tmp_path / text_file_name
    test_file.write_text(text_file_content)

    # Test using the FOLDER path
    valid_upload_item = {"path": str(tmp_path), "stub": True}

    validate_upload_item(valid_upload_item)


def test_validate_upload_item_validate_dictionary():
    """Test that validate_upload_item raises a TypeError when upload_item is
    not a dictionary.
    """

    upload_item = 5
    with pytest.raises(
        TypeError,
        match="'upload_item' must be a dictionary: 'upload_item = 5'",
    ):
        validate_upload_item(upload_item)


def test_derive_integrity_data_file_doesnt_exist():
    """Test that derive_integrity_data raises a FileExistsError when using
    a file path that does not exist.
    """

    file_path = "f a k e f i l e p a t h"
    with pytest.raises(
        FileExistsError, match=f"'file_path' does not exist: '{file_path}'"
    ):
        derive_integrity_data(file_path)


def test_derive_integrity_data_path_is_string():
    """Test that derive_integrity_data raises a TypeError when the file path is
    not a string.
    """

    file_path = 10
    with pytest.raises(
        TypeError, match=f"'file_path' must be a string: '{file_path}'"
    ):
        derive_integrity_data(file_path)


def test_derive_integrity_data_byte_count_string(tmp_path):
    """Test that derive_integrity_data raises a TypeError when byte count is
    not an integer.
    """

    text_file_name = "test.txt"
    text_file_content = "0123456789"

    test_file = tmp_path / text_file_name
    test_file.write_text(text_file_content)
    byte_count_in = "invalid byte count"

    with pytest.raises(
        TypeError, match=f"'byte_count' must be an int: '{byte_count_in}'"
    ):
        derive_integrity_data(str(test_file), byte_count=byte_count_in)


def test_derive_integrity_data_byte_count_float(tmp_path):
    """Test that derive_integrity_data raises a TypeError when byte count is
    not an integer.
    """

    text_file_name = "test.txt"
    text_file_content = "0123456789"

    test_file = tmp_path / text_file_name
    test_file.write_text(text_file_content)
    byte_count_in = 6.5

    with pytest.raises(
        TypeError, match=f"'byte_count' must be an int: '{byte_count_in}'"
    ):
        derive_integrity_data(str(test_file), byte_count=byte_count_in)


def test_derive_integrity_data_byte_count_0(tmp_path):
    """Test that derive_integrity_data raises a ValueError when byte count is
    not greater than 0.
    """

    text_file_name = "test.txt"
    text_file_content = "0123456789"

    test_file = tmp_path / text_file_name
    test_file.write_text(text_file_content)
    byte_count_in = 0

    with pytest.raises(
        ValueError,
        match=f"'byte_count' must be greater than zero: '{byte_count_in}'",
    ):
        derive_integrity_data(str(test_file), byte_count=byte_count_in)


def test_derive_integrity_data_large_byte_count(tmp_path):
    """Test that derive_integrity_data returns the expected hash values
    when using large byte counts.
    """

    text_file_name = "test.txt"
    text_file_content = "0123456789"

    test_file = tmp_path / text_file_name
    test_file.write_text(text_file_content)

    # Expected sha256 hash of the string "0123456789". The whole file is used
    # as the file size is > 3*byte_count.
    e_hash = "84d89877f0d4041efb6bf91a16f0248f2fd573e6af05c19f96bedb9f882f7882"
    assert e_hash == derive_integrity_data(str(test_file), 10**6)["hash"]


def test_derive_integrity_data_small_byte_count(tmp_path):
    """Test that derive_integrity_data returns the expected hash values when
    using small byte counts.
    """

    text_file_name = "test.txt"
    text_file_content = "0123456789"

    test_file = tmp_path / text_file_name
    test_file.write_text(text_file_content)

    # Expected sha256 hash of the string "014589". This string is used as the
    # file size is <= 3*byte_count. So checksum is generated using bytes from
    # beginning, middle, and end (instead of whole file content).
    e_hash = "a2a7cb1d7fc8f79e33b716b328e19bb381c3ec96a2dca02a3d1183e7231413bb"
    assert e_hash == derive_integrity_data(str(test_file), 2)["hash"]


def test_derive_integrity_data_file_size(tmp_path):
    """Test that derive_integrity_data returns the correct file size value."""
    text_file_name = "test.txt"
    text_file_content = "0123456789"

    test_file = tmp_path / text_file_name
    test_file.write_text(text_file_content)

    # Check the size of the file is accurate to the length of it's contents.
    assert derive_integrity_data(str(test_file), 2)["file_size"] == len(
        text_file_content
    )


def test_derive_integrity_data_date(tmp_path):
    """Test that derive_integrity_data returns the correct data access date."""
    text_file_name = "test.txt"
    text_file_content = "0123456789"

    test_file = tmp_path / text_file_name
    test_file.write_text(text_file_content)

    access_date = derive_integrity_data(str(test_file), 2)["access_date"]
    access_date = datetime.fromisoformat(access_date)
    access_date_seconds = access_date.timestamp()
    now_seconds = time.time_ns() // 10**9
    diff_seconds = abs(access_date_seconds - now_seconds)
    assert diff_seconds < 60


def test_derive_integrity_data_source_path(tmp_path):
    """Test that derive_integrity_data returns the correct source path value.
    """

    text_file_name = "test.txt"
    text_file_content = "0123456789"

    test_file = tmp_path / text_file_name
    test_file.write_text(text_file_content)

    assert derive_integrity_data(str(test_file), 2)["source_path"] == str(
        test_file
    )


def test_derive_integrity_byte_count(tmp_path):
    """Test that derive_integrity_data returns the correct byte count value."""
    text_file_name = "test.txt"
    text_file_content = "0123456789"

    test_file = tmp_path / text_file_name
    test_file.write_text(text_file_content)

    assert derive_integrity_data(str(test_file), 2)["byte_count"] == 2


def test_derive_integrity_data_max_byte_count_off_by_one(tmp_path):
    """Test that derive_integrity_data raises a ValueError when the passed in
    byte count is > 300,000,000 bytes.
    """

    text_file_name = "test.txt"
    text_file_content = "0123456789"

    test_file = tmp_path / text_file_name
    test_file.write_text(text_file_content)

    byte_count_in = 300 * 10**6 + 1

    with pytest.raises(
        ValueError,
        match=f"'byte_count' must be less "
        f"than or equal to 300000000 bytes: '{byte_count_in}'",
    ):
        derive_integrity_data(str(test_file), byte_count=byte_count_in)


def test_derive_integrity_data_max_byte_count_exact(tmp_path):
    """Test that derive_integrity_data runs successfully when the passed in
    byte count is exactly 300,000,000 bytes.
    """

    text_file_name = "test.txt"
    text_file_content = "0123456789"

    test_file = tmp_path / text_file_name
    test_file.write_text(text_file_content)

    byte_count_in = 300 * 10**6 + 1

    derive_integrity_data(str(test_file), byte_count=byte_count_in - 1)


# Test with two different fsspec file systems (top of file).
@pytest.fixture(params=[s3fs, local_fs])
def test_basket(request, tmpdir):
    """Sets up pytest fixture."""
    file_system = request.param
    test_pantry = PantryForTest(tmpdir, file_system)
    yield test_pantry
    test_pantry.cleanup_pantry()


def test_upload_basket_without_uuid_creates_uuid(test_basket):
    """Test that upload_basket creates a uuid when unique_id is not
    initialized.
    """

    # Create a temporary basket with a test file.
    tmp_basket_dir_name = "test_basket_tmp_dir"
    tmp_dir = test_basket.set_up_basket(tmp_basket_dir_name)

    #Initialize all kwargs except unique_id
    upload_items = [{"path": tmp_dir.strpath, "stub": False}]
    basket_type = "test_basket"
    upload_path = os.path.join(test_basket.pantry_path, basket_type)

    uploading_basket = weave.upload.UploadBasket(
        upload_items=upload_items,
        upload_directory=upload_path,
        basket_type=basket_type,
        file_system=test_basket.file_system,
    )

    assert uploading_basket.kwargs.get("unique_id") is not None

    tmp_files = test_basket.file_system.ls(upload_path)
    manifest_path = [s for s in tmp_files if s.endswith('manifest.json')][0]
    with test_basket.file_system.open(manifest_path, "r", encoding="utf-8")\
            as outfile:
        manifest_data = json.load(outfile)

    assert manifest_data["uuid"] != "null"


def test_upload_basket_upload_items_is_not_a_string(test_basket):
    """Test that upload_basket raises a TypeError when upload_items is not a
    list of dictionaries.
    """

    # Create a temporary basket with a test file.
    tmp_basket_dir_name = "test_basket_tmp_dir"
    test_basket.set_up_basket(tmp_basket_dir_name)

    upload_items = "n o t a r e a l p a t h"
    unique_id = uuid.uuid1().hex
    basket_type = "test_basket"
    upload_path = os.path.join(test_basket.pantry_path, basket_type, unique_id)

    with pytest.raises(
        TypeError,
        match="'upload_items' must be a list of "
        + f"dictionaries: '{upload_items}'",
    ):
        UploadBasket(
            upload_items=upload_items,
            upload_directory=upload_path,
            unique_id=unique_id,
            basket_type=basket_type,
            file_system=test_basket.file_system,
        )

    assert not test_basket.file_system.exists(upload_path)


def test_upload_basket_upload_items_is_not_a_list_of_strings(test_basket):
    """Test that upload_basket raises a TypeError when upload_items is not a
    list of dictionaries.
    """

    # Create a temporary basket with a test file.
    tmp_basket_dir_name = "test_basket_tmp_dir"
    test_basket.set_up_basket(tmp_basket_dir_name)

    upload_items = ["invalid", "invalid2"]
    unique_id = uuid.uuid1().hex
    basket_type = "test_basket"
    upload_path = os.path.join(test_basket.pantry_path, basket_type, unique_id)

    with pytest.raises(
        TypeError, match="'upload_items' must be a list of dictionaries:"
    ):
        UploadBasket(
            upload_items=upload_items,
            upload_directory=upload_path,
            unique_id=unique_id,
            basket_type=basket_type,
            file_system=test_basket.file_system,
        )

    assert not test_basket.file_system.exists(upload_path)


def test_upload_basket_upload_items_is_a_list_of_only_dictionaries(
    test_basket,
):
    """Test that upload_basket raises a TypeError when upload_items is not a
    list of dictionaries.
    """

    # Create a temporary basket with a test file.
    tmp_basket_dir_name = "test_basket_tmp_dir"
    test_basket.set_up_basket(tmp_basket_dir_name)

    upload_items = [{}, "invalid2"]
    unique_id = uuid.uuid1().hex
    basket_type = "test_basket"
    upload_path = os.path.join(test_basket.pantry_path, basket_type, unique_id)

    with pytest.raises(
        TypeError, match="'upload_items' must be a list of dictionaries:"
    ):
        UploadBasket(
            upload_items=upload_items,
            upload_directory=upload_path,
            unique_id=unique_id,
            basket_type=basket_type,
            file_system=test_basket.file_system,
        )

    assert not test_basket.file_system.exists(upload_path)


@patch("weave.upload.UploadBasket.upload_basket_supplement_to_fs")
def test_upload_basket_with_bad_upload_items_is_deleted_if_it_fails(
    mocked_obj, test_basket
):
    """Test that upload_basket deletes bad upload items if it fails to upload.
    """

    # Create a temporary basket with a test file.
    tmp_basket_dir_name = "test_basket_tmp_dir"
    tmp_dir = test_basket.set_up_basket(tmp_basket_dir_name)

    upload_items = [{"path": tmp_dir.strpath, "stub": False}]
    unique_id = uuid.uuid1().hex
    basket_type = "test_basket"
    upload_path = os.path.join(test_basket.pantry_path, basket_type, unique_id)

    with pytest.raises(
        ValueError,
        match="This error provided for"
        "test_upload_basket_with_bad_upload_items_is_deleted_if_it_fails",
    ):
        mocked_obj.side_effect = ValueError(
            "This error provided for"
            "test_upload_basket_with_bad_upload_items_is_deleted_if_it_fails"
        )
        UploadBasket(
            upload_items=upload_items,
            upload_directory=upload_path,
            unique_id=unique_id,
            basket_type=basket_type,
            file_system=test_basket.file_system,
        )

    assert not test_basket.file_system.exists(upload_path)


def test_upload_basket_upload_items_invalid_dictionary(test_basket):
    """Test that upload_basket raises a KeyError when upload_items contains an
    invalid path key.
    """

    # Create a temporary basket with a test file.
    tmp_basket_dir_name = "test_basket_tmp_dir"
    tmp_basket_txt_file_name = "test.txt"
    tmp_basket_dir = test_basket.set_up_basket(
        tmp_basket_dir_name, file_name=tmp_basket_txt_file_name
    )
    tmp_basket_txt_file = tmp_basket_dir.join(tmp_basket_txt_file_name)

    unique_id = uuid.uuid1().hex
    basket_type = "test_basket"
    upload_path = os.path.join(test_basket.pantry_path, basket_type, unique_id)

    upload_items = [
        {
            "path": tmp_basket_dir.strpath,
            "stub": True,
        },
        {"path_invalid_key": tmp_basket_txt_file.strpath, "stub": True},
    ]
    with pytest.raises(
        KeyError, match="Invalid upload_item key: 'path_invalid_key'"
    ):
        UploadBasket(
            upload_items=upload_items,
            upload_directory=upload_path,
            unique_id=unique_id,
            basket_type=basket_type,
            file_system=test_basket.file_system,
        )


def test_upload_basket_upload_items_check_unique_file_folder_names(
    test_basket,
):
    """Test that upload_basket raises ValueErrors when upload_items does not
    contain unique file and folder names.
    """

    unique_id = uuid.uuid1().hex
    basket_type = "test_basket"
    tmp_basket_txt_file_name = "test.txt"

    # Create a temporary basket with a test file.
    tmp_basket_dir_name1 = "test_basket_tmp_dir"
    tmp_basket_dir1 = test_basket.set_up_basket(
        tmp_basket_dir_name1, file_name=tmp_basket_txt_file_name
    )
    tmp_basket_txt_file1 = tmp_basket_dir1.join(tmp_basket_txt_file_name)

    # Manually add another test file with the same name as the other.
    tmp_basket_dir_name2 = "test_basket_tmp_dir2"
    tmp_basket_dir2 = test_basket.set_up_basket(
        tmp_basket_dir_name2, file_name=tmp_basket_txt_file_name
    )
    tmp_basket_txt_file2 = tmp_basket_dir2.join(tmp_basket_txt_file_name)

    upload_path = os.path.join(test_basket.pantry_path, basket_type, unique_id)

    # Test same file names
    upload_items = [
        {"path": tmp_basket_txt_file1.strpath, "stub": True},
        {"path": tmp_basket_txt_file2.strpath, "stub": True},
    ]
    with pytest.raises(
        ValueError,
        match="'upload_item' folder and file names must be unique:"
        " Duplicate Name = test.txt",
    ):
        UploadBasket(
            upload_items=upload_items,
            upload_directory=upload_path,
            unique_id=unique_id,
            basket_type=basket_type,
            file_system=test_basket.file_system,
        )

    # Test same dirname
    upload_items = [
        {"path": tmp_basket_dir1.strpath, "stub": True},
        {"path": tmp_basket_dir1.strpath, "stub": True},
    ]
    with pytest.raises(
        ValueError,
        match="'upload_item' folder and file names must be unique:"
        " Duplicate Name = test_basket_tmp_dir",
    ):
        UploadBasket(
            upload_items=upload_items,
            upload_directory=upload_path,
            unique_id=unique_id,
            basket_type=basket_type,
            file_system=test_basket.file_system,
        )

    # Test same dirname same file
    upload_items = [
        {"path": tmp_basket_txt_file1.strpath, "stub": True},
        {"path": tmp_basket_txt_file1.strpath, "stub": True},
    ]
    with pytest.raises(
        ValueError,
        match="'upload_item' folder and file names must be unique:"
        " Duplicate Name = test.txt",
    ):
        UploadBasket(
            upload_items=upload_items,
            upload_directory=upload_path,
            unique_id=unique_id,
            basket_type=basket_type,
            file_system=test_basket.file_system,
        )

    assert not test_basket.file_system.exists(f"{upload_path}")


def test_upload_basket_upload_path_is_string(test_basket):
    """Test that upload_basket raises a TypeError when upload_items is not a
    list of dictionaries.
    """

    # Create a temporary basket with a test file.
    tmp_basket_dir_name = "test_basket_tmp_dir"
    tmp_basket_dir = test_basket.set_up_basket(tmp_basket_dir_name)

    upload_items = [
        {
            "path": tmp_basket_dir.strpath,
            "stub": True,
        }
    ]

    basket_type = "test_basket"
    unique_id = uuid.uuid1().hex
    upload_path = 1234

    with pytest.raises(
        TypeError,
        match="Invalid datatype: 'upload_directory: "
        "must be type <class 'str'>'",
    ):
        UploadBasket(
            upload_items=upload_items,
            upload_directory=upload_path,
            unique_id=unique_id,
            basket_type=basket_type,
            file_system=test_basket.file_system,
        )


def test_upload_basket_unique_id_string(test_basket):
    """Test that upload_basket raises a TypeError when unique id is not a
    string.
    """

    # Create a temporary basket with a test file.
    tmp_basket_dir_name = "test_basket_tmp_dir"
    tmp_basket_dir = test_basket.set_up_basket(tmp_basket_dir_name)

    upload_items = [
        {
            "path": tmp_basket_dir.strpath,
            "stub": True,
        }
    ]

    basket_type = "test_basket"
    unique_id = 6
    upload_path = os.path.join(
        test_basket.pantry_path, basket_type, f"{unique_id}"
    )

    with pytest.raises(
        TypeError,
        match="Invalid datatype: 'unique_id: must be " "type <class 'str'>'",
    ):
        UploadBasket(
            upload_items=upload_items,
            upload_directory=upload_path,
            unique_id=unique_id,
            basket_type=basket_type,
            file_system=test_basket.file_system,
        )

    assert not test_basket.file_system.exists(f"{upload_path}")


def test_upload_basket_type_is_string(test_basket):
    """Test that upload_basket raises TypeError when basket type is not a
    string.
    """

    # Create a temporary basket with a test file.
    tmp_basket_dir_name = "test_basket_tmp_dir"
    tmp_basket_dir = test_basket.set_up_basket(tmp_basket_dir_name)

    upload_items = [
        {
            "path": tmp_basket_dir.strpath,
            "stub": True,
        }
    ]

    basket_type = 1234
    unique_id = uuid.uuid1().hex
    upload_path = os.path.join(
        test_basket.pantry_path, f"{basket_type}", unique_id
    )

    with pytest.raises(
        TypeError,
        match="Invalid datatype: 'basket_type: must be type <class 'str'>'",
    ):
        UploadBasket(
            upload_items=upload_items,
            upload_directory=upload_path,
            unique_id=unique_id,
            basket_type=basket_type,
            file_system=test_basket.file_system,
        )

    assert not test_basket.file_system.exists(f"{upload_path}")


def test_upload_basket_parent_ids_list_str(test_basket):
    """Test that upload_basket raises a TypeError when parent ids is not a list
    of strings.
    """

    # Create a temporary basket with a test file.
    tmp_basket_dir_name = "test_basket_tmp_dir"
    tmp_basket_dir = test_basket.set_up_basket(tmp_basket_dir_name)

    upload_items = [
        {
            "path": tmp_basket_dir.strpath,
            "stub": True,
        }
    ]

    basket_type = "test_basket"
    unique_id = uuid.uuid1().hex
    upload_path = os.path.join(
        test_basket.pantry_path, f"{basket_type}", unique_id
    )
    parent_ids_in = ["a", 3]

    with pytest.raises(
        TypeError, match="'parent_ids' must be a list of strings:"
    ):
        UploadBasket(
            upload_items=upload_items,
            upload_directory=upload_path,
            unique_id=unique_id,
            basket_type=basket_type,
            parent_ids=parent_ids_in,
            file_system=test_basket.file_system,
        )

    assert not test_basket.file_system.exists(f"{upload_path}")


def test_upload_basket_parent_ids_is_list(test_basket):
    """Test that upload_basket raises a TypeError when parent ids is
    not a list.
    """

    # Create a temporary basket with a test file.
    tmp_basket_dir_name = "test_basket_tmp_dir"
    tmp_basket_dir = test_basket.set_up_basket(tmp_basket_dir_name)

    upload_items = [
        {
            "path": tmp_basket_dir.strpath,
            "stub": True,
        }
    ]

    basket_type = "test_basket"
    unique_id = uuid.uuid1().hex
    upload_path = os.path.join(
        test_basket.pantry_path, f"{basket_type}", unique_id
    )
    parent_ids_in = 56

    with pytest.raises(
        TypeError,
        match="Invalid datatype: 'parent_ids: must be type <class 'list'>'",
    ):
        UploadBasket(
            upload_items=upload_items,
            upload_directory=upload_path,
            unique_id=unique_id,
            basket_type=basket_type,
            parent_ids=parent_ids_in,
            file_system=test_basket.file_system,
        )

    assert not test_basket.file_system.exists(f"{upload_path}")


def test_upload_basket_metadata_is_dictionary(test_basket):
    """Test that upload_basket raises a TypeError when metadata is not a
    dictionary.
    """

    # Create a temporary basket with a test file.
    tmp_basket_dir_name = "test_basket_tmp_dir"
    tmp_basket_dir = test_basket.set_up_basket(tmp_basket_dir_name)

    upload_items = [
        {
            "path": tmp_basket_dir.strpath,
            "stub": True,
        }
    ]

    basket_type = "test_basket"
    unique_id = uuid.uuid1().hex
    upload_path = os.path.join(
        test_basket.pantry_path, f"{basket_type}", unique_id
    )
    metadata_in = "invalid"

    with pytest.raises(
        TypeError,
        match="Invalid datatype: 'metadata: must be type <class 'dict'>'",
    ):
        UploadBasket(
            upload_items=upload_items,
            upload_directory=upload_path,
            unique_id=unique_id,
            basket_type=basket_type,
            metadata=metadata_in,
            file_system=test_basket.file_system,
        )

    assert not test_basket.file_system.exists(f"{upload_path}")


def test_upload_basket_label_is_string(test_basket):
    """Test that upload_basket raises a TypeError when the label is
    not a string.
    """

    # Create a temporary basket with a test file.
    tmp_basket_dir_name = "test_basket_tmp_dir"
    tmp_basket_dir = test_basket.set_up_basket(tmp_basket_dir_name)

    upload_items = [
        {
            "path": tmp_basket_dir.strpath,
            "stub": True,
        }
    ]

    basket_type = "test_basket"
    unique_id = uuid.uuid1().hex
    upload_path = os.path.join(
        test_basket.pantry_path, f"{basket_type}", unique_id
    )
    label_in = 1234

    with pytest.raises(
        TypeError,
        match="Invalid datatype: 'label: must be type <class 'str'>'",
    ):
        UploadBasket(
            upload_items=upload_items,
            upload_directory=upload_path,
            unique_id=unique_id,
            basket_type=basket_type,
            label=label_in,
            file_system=test_basket.file_system,
        )

    assert not test_basket.file_system.exists(f"{upload_path}")


def test_upload_basket_no_metadata(test_basket):
    """Test that no metadata is created if no metadata is passed to 
    upload_pantry.
    """

    # Create a temporary basket with a test file.
    tmp_basket_dir_name = "test_basket_tmp_dir"
    tmp_basket_dir = test_basket.set_up_basket(tmp_basket_dir_name)

    upload_items = [
        {
            "path": tmp_basket_dir.strpath,
            "stub": True,
        }
    ]

    basket_type = "test_basket"
    unique_id = uuid.uuid1().hex
    upload_path = os.path.join(
        test_basket.pantry_path, f"{basket_type}", unique_id
    )

    UploadBasket(
        upload_items=upload_items,
        upload_directory=upload_path,
        unique_id=unique_id,
        basket_type=basket_type,
        file_system=test_basket.file_system,
    )

    # Assert metadata.json was not written
    assert not test_basket.file_system.exists(f"{upload_path}/metadata.json")


def test_upload_basket_check_existing_upload_path(test_basket):
    """Test that upload_basket raises a FileExistsError when the upload
    directory already exists.
    """

    # Create a temporary basket with a test file.
    tmp_basket_dir_name = "test_basket_tmp_dir"
    tmp_basket_dir = test_basket.set_up_basket(tmp_basket_dir_name)

    upload_items = [
        {
            "path": tmp_basket_dir.strpath,
            "stub": True,
        }
    ]

    basket_type = "test_basket"
    unique_id = uuid.uuid1().hex
    upload_path = os.path.join(
        test_basket.pantry_path, f"{basket_type}", unique_id
    )

    test_basket.file_system.upload(
        tmp_basket_dir.strpath, f"{upload_path}", recursive=True
    )

    with pytest.raises(
        FileExistsError,
        match=f"'upload_directory' already exists: '{upload_path}''",
    ):
        UploadBasket(
            upload_items=upload_items,
            upload_directory=upload_path,
            unique_id=unique_id,
            basket_type=basket_type,
            file_system=test_basket.file_system,
        )

    assert test_basket.file_system.ls(
        os.path.join(test_basket.pantry_path, f"{basket_type}")
    )[0].endswith(upload_path)


def test_upload_basket_check_unallowed_file_names(test_basket):
    """Test that upload_basket raises a ValueError when trying to upload files
    with reserved/unallowed file names.
    """

    json_data = {"t": [1, 2, 3]}
    unallowed_file_names = [
        "basket_manifest.json",
        "basket_metadata.json",
        "basket_supplement.json",
    ]
    for ind, unallowed_file_name in enumerate(unallowed_file_names):
        # Create a temporary basket with a test file.
        tmp_basket_dir_name = f"test_basket_tmp_dir{ind}"
        tmp_basket_dir = test_basket.set_up_basket(
            tmp_basket_dir_name,
            file_name=unallowed_file_name,
            file_content=json_data,
        )

        unallowed_file_path = tmp_basket_dir.join(unallowed_file_name)

        upload_items = [
            {
                "path": unallowed_file_path.strpath,
                "stub": True,
            }
        ]

        basket_type = "test_basket"
        unique_id = uuid.uuid1().hex
        upload_path = os.path.join(
            test_basket.pantry_path, f"{basket_type}", unique_id
        )

        with pytest.raises(
            ValueError,
            match=f"'{unallowed_file_name}' filename not allowed",
        ):
            UploadBasket(
                upload_items=upload_items,
                upload_directory=upload_path,
                unique_id=unique_id,
                basket_type=basket_type,
                file_system=test_basket.file_system,
            )

    assert not test_basket.file_system.exists(f"{upload_path}")


def test_upload_basket_clean_up_on_error(test_basket):
    """Test that upload_basket cleans up failed basket uploads when any
    Exception is encountered and the test_cleanup_flag is passed.
    """

    # Create a temporary basket with a test file.
    tmp_basket_dir_name = "test_basket_tmp_dir"
    tmp_basket_dir = test_basket.set_up_basket(tmp_basket_dir_name)

    upload_items = [
        {
            "path": tmp_basket_dir.strpath,
            "stub": True,
        }
    ]

    basket_type = "test_basket"
    unique_id = uuid.uuid1().hex
    upload_path = os.path.join(
        test_basket.pantry_path, f"{basket_type}", unique_id
    )

    with pytest.raises(Exception, match="Test Clean Up"):
        UploadBasket(
            upload_items=upload_items,
            upload_directory=upload_path,
            unique_id=unique_id,
            basket_type=basket_type,
            file_system=test_basket.file_system,
            test_clean_up=True,
        )

    assert not test_basket.file_system.exists(upload_path)


def test_upload_basket_invalid_optional_argument(test_basket):
    """Test that upload_basket raises a KeyError when an invalid optional
    argument is passed.
    """

    # Create a temporary basket with a test file.
    tmp_basket_dir_name = "test_basket_tmp_dir"
    tmp_basket_dir = test_basket.set_up_basket(tmp_basket_dir_name)

    upload_items = [
        {
            "path": tmp_basket_dir.strpath,
            "stub": True,
        }
    ]

    basket_type = "test_basket"
    unique_id = uuid.uuid1().hex
    upload_path = os.path.join(
        test_basket.pantry_path, f"{basket_type}", unique_id
    )

    with pytest.raises(KeyError, match="Invalid kwargs argument: 'junk'"):
        UploadBasket(
            upload_items=upload_items,
            upload_directory=upload_path,
            unique_id=unique_id,
            basket_type=basket_type,
            file_system=test_basket.file_system,
            junk=True,
        )

    assert not test_basket.file_system.exists(upload_path)


def test_upload_basket_invalid_test_clean_up_datatype(test_basket):
    """Test that upload_basket raises a TypeError when the optional
    test_clean_up argument is not a bool.
    """

    # Create a temporary basket with a test file.
    tmp_basket_dir_name = "test_basket_tmp_dir"
    tmp_basket_dir = test_basket.set_up_basket(tmp_basket_dir_name)

    upload_items = [
        {
            "path": tmp_basket_dir.strpath,
            "stub": True,
        }
    ]

    basket_type = "test_basket"
    unique_id = uuid.uuid1().hex
    upload_path = os.path.join(
        test_basket.pantry_path, f"{basket_type}", unique_id
    )

    with pytest.raises(
        TypeError,
        match="Invalid datatype: 'test_clean_up: must be type <class 'bool'>'",
    ):
        UploadBasket(
            upload_items=upload_items,
            upload_directory=upload_path,
            unique_id=unique_id,
            basket_type=basket_type,
            file_system=test_basket.file_system,
            test_clean_up="a",
        )

    assert not test_basket.file_system.exists(upload_path)


def test_upload_basket_file_contents_identical(test_basket):
    """Test that files uploaded using upload_basket are the same as local."""
    # Create a temporary basket with a test file.
    test_file_name = "test.txt"
    tmp_basket_dir_name = "test_basket_tmp_dir"
    tmp_basket_dir = test_basket.set_up_basket(
        tmp_basket_dir_name, file_name=test_file_name
    )
    tmp_basket_file_path = tmp_basket_dir.join(test_file_name)

    upload_items = [
        {
            "path": tmp_basket_dir.strpath,
            "stub": False,
        }
    ]

    basket_type = "test_basket"
    unique_id = uuid.uuid1().hex
    upload_path = os.path.join(
        test_basket.pantry_path, f"{basket_type}", unique_id
    )
    upload_file_path = os.path.join(
        upload_path, tmp_basket_dir_name, test_file_name
    )

    UploadBasket(
        upload_items=upload_items,
        upload_directory=upload_path,
        unique_id=unique_id,
        basket_type=basket_type,
        file_system=test_basket.file_system,
    )

    # Read the file data, then assert the uploaded contents are the same.
    with open(tmp_basket_file_path, "r", encoding="UTF-8") as r_file:
        local_file_data = r_file.read()

    with test_basket.file_system.open(upload_file_path, "r") as r_file:
        assert r_file.read() == local_file_data


def test_upload_correct_version_number(test_basket):
    """Test that when a basket is uploaded, the manifest contains the
    correct version of weave.
    """

    tmp_basket_dir_name = "test_basket_tmp_dir"
    tmp_basket_dir = test_basket.set_up_basket(tmp_basket_dir_name)
    upload_path = test_basket.upload_basket(tmp_basket_dir)

    manifest_path = os.path.join(upload_path, "basket_manifest.json")
    with test_basket.file_system.open(manifest_path, "r") as file:
        manifest_dict = json.load(file)

    assert manifest_dict["weave_version"] == weave.__version__


def test_upload_metadata_only_basket(test_basket):
    """Try to upload a valid metadata-only basket
    """
    pantry = Pantry(IndexPandas,
                    pantry_path=test_basket.pantry_path,
                    file_system=test_basket.file_system)

    basket = pantry.upload_basket(upload_items=[],
                         basket_type="metadata_only",
                         parent_ids=["1"],
                         metadata={"metadata":"only"})

    assert len(basket) == 1


def test_upload_basket_no_files(test_basket):
    """Upload a basket with no files included, ensure an error is thrown.
    """
    pantry = Pantry(IndexPandas,
                    pantry_path=test_basket.pantry_path,
                    file_system=test_basket.file_system)

    with pytest.raises(
        ValueError,
        match=(r"Files are required to upload a basket. If you want a metadata"
               r"-only basket, please include metadata and parent uid\(s\)"),
    ):
        pantry.upload_basket(upload_items=[], basket_type="no_files",)

def test_upload_from_s3fs(test_basket):
    """Test that a basket can be uploaded from s3fs to the local file
    system or s3fs.
    """
    pantry_path = os.path.join(test_basket.pantry_path, "test-pantry-1")

    test_basket.file_system.mkdir(pantry_path)

    tmp_txt_file = test_basket.tmpdir.join("test.txt")
    tmp_txt_file.write("this is a test")

    nested_path = os.path.join(pantry_path, "text.txt")
    file_to_move = os.path.join(test_basket.pantry_path, "text.txt")

    if test_basket.file_system == local_fs:
        minio_path = "test-source-file-system"
        try:
            s3fs.mkdir(minio_path)
        except FileExistsError:
            s3fs.rm(minio_path,recursive=True)
            s3fs.mkdir(minio_path)
        file_to_move = os.path.join(minio_path, "test.txt")

    # Must upload a file because Minio will remove empty directories
    if test_basket.file_system == local_fs:
        s3fs.upload(str(tmp_txt_file.realpath()), file_to_move)
    else:
        test_basket.file_system.upload(str(tmp_txt_file.realpath()),
                                       file_to_move)
    test_basket.file_system.upload(str(tmp_txt_file.realpath()),
                                   nested_path)
    # Make the Pantries.
    pantry_1 = Pantry(
        IndexPandas,
        pantry_path=pantry_path,
        file_system=test_basket.file_system,
    )
    pantry_2 = Pantry(
        IndexSQLite,
        pantry_path=pantry_path,
        file_system=test_basket.file_system,
    )
    basket_uuid = pantry_1.upload_basket(
                           upload_items=[{'path': file_to_move,'stub':False}],
                           source_file_system=s3fs,
                           basket_type="test_s3fs_upload"
                           )["uuid"][0]
    basket_uuid2 = pantry_2.upload_basket(
                       upload_items=[{'path': file_to_move,'stub':False}],
                       source_file_system=s3fs,
                       basket_type="test_s3fs_upload"
                       )["uuid"][0]
    basket = pantry_1.get_basket(basket_uuid)
    basket2 = pantry_2.get_basket(basket_uuid2)
    if not test_basket.file_system == local_fs:
        assert (basket.ls()[0].endswith('text.txt') and
                basket2.ls()[0].endswith('text.txt'))
    else:
        assert (basket.ls()[0].endswith('test.txt') and
                basket2.ls()[0].endswith('test.txt'))
        s3fs.rm(minio_path,recursive=True)
        remove_path = pantry_2.index.db_path
        if local_fs.exists(remove_path):
            local_fs.rm(remove_path)



"""Pytests for the metadata_db functionality."""

import os
import sys

import pandas as pd
import pytest
import s3fs
from fsspec.implementations.local import LocalFileSystem

import weave
from weave.tests.pytest_resources import PantryForTest


class MongoForTest(PantryForTest):
    """Extend the PantryForTest class to support mongodb and custom data
    loader.
    """

    def __init__(self, tmpdir, file_system):
        super().__init__(tmpdir, file_system)
        self.test_collection = "test_collection"
        self.mongodb = weave.config.get_mongo_db().mongo_metadata
        self.load_data()

    def load_data(self):
        """Loads data into the file_system."""
        # Create a temporary basket with a test file.
        tmp_basket_dir_name = "test_basket_tmp_dir"
        tmp_basket_dir = self.set_up_basket(tmp_basket_dir_name)

        # Upload the basket with different uuids and metadata.
        self.upload_basket(
            tmp_basket_dir, uid="1234", metadata={"key1": "value1"}
        )

        tmp_nested_dir = self.add_lower_dir_to_temp_basket(tmp_basket_dir)
        self.upload_basket(
            tmp_nested_dir, uid="4321", metadata={"key2": "value2"}
        )

        self.upload_basket(tmp_basket_dir, uid="nometadata")

    def cleanup(self):
        """Cleans up the pantry and mongodb."""
        self.cleanup_pantry()
        self.mongodb[self.test_collection].drop()


s3fs = s3fs.S3FileSystem(
    client_kwargs={"endpoint_url": os.environ["S3_ENDPOINT"]}
)
local_fs = LocalFileSystem()


# Test with two different fsspec file systems (above).
@pytest.fixture(
    params=[s3fs, local_fs],
    ids=["S3FileSystem", "LocalFileSystem"],
)
def set_up(request, tmpdir):
    """Sets up the fixture for testing usage."""
    file_system = request.param
    database = MongoForTest(tmpdir, file_system)
    yield database
    database.cleanup()


# Ignoring pylint's warning "redefined-outer-name" as this is simply
# how pytest works when it comes to pytest fixtures.
# pylint: disable=redefined-outer-name

# Skip tests if pymongo is not installed.
@pytest.mark.skipif(
    "pymongo" not in sys.modules or not os.environ.get("MONGODB_HOST", False),
    reason="Pymongo required for this test",
)
def test_load_mongo(set_up):
    """Test that load_mongo successfully loads valid metadata to the set_up.
    """

    index_table = weave.index.create_index.create_index_from_fs(
        set_up.pantry_path, set_up.file_system
    )
    weave.load_mongo(
        index_table,
        file_system=set_up.file_system,
        collection=set_up.test_collection,
    )

    truth_db = [
        {"uuid": "1234", "basket_type": "test_basket", "key1": "value1"},
        {"uuid": "4321", "basket_type": "test_basket", "key2": "value2"},
    ]

    db_data = list(set_up.mongodb[set_up.test_collection].find({}))
    compared_data = []
    for item in db_data:
        item.pop("_id")
        compared_data.append(item)
    assert truth_db == compared_data


@pytest.mark.skipif(
    "pymongo" not in sys.modules or not os.environ.get("MONGODB_HOST", False),
    reason="Pymongo required for this test",
)
def test_load_mongo_check_for_dataframe(set_up):
    """Test that load_mongo prevents loading data with an invalid index_table.
    """

    with pytest.raises(
        TypeError,
        match="Invalid datatype for index_table: " "must be Pandas DataFrame",
    ):
        weave.load_mongo(
            "",
            file_system=set_up.file_system,
            collection=set_up.test_collection,
        )


@pytest.mark.skipif(
    "pymongo" not in sys.modules or not os.environ.get("MONGODB_HOST", False),
    reason="Pymongo required for this test",
)
def test_load_mongo_check_collection_for_string(set_up):
    """Test that load_mongo prevents loading data with an invalid set_up
    collection.
    """

    with pytest.raises(
        TypeError, match="Invalid datatype for collection: " "must be a string"
    ):
        weave.load_mongo(
            pd.DataFrame(), file_system=set_up.file_system, collection=1
        )


@pytest.mark.skipif(
    "pymongo" not in sys.modules or not os.environ.get("MONGODB_HOST", False),
    reason="Pymongo required for this test",
)
def test_load_mongo_check_dataframe_for_uuid(set_up):
    """Test that load_mongo prevents loading data with missing uuid.
    """

    with pytest.raises(
        ValueError, match="Invalid index_table: " "missing uuid column"
    ):
        weave.load_mongo(
            pd.DataFrame({"basket_type": ["type"], "address": ["path"]}),
            file_system=set_up.file_system,
            collection=set_up.test_collection,
        )


@pytest.mark.skipif(
    "pymongo" not in sys.modules or not os.environ.get("MONGODB_HOST", False),
    reason="Pymongo required for this test",
)
def test_load_mongo_check_dataframe_for_address(set_up):
    """Test that load_mongo prevents loading data with missing address."""
    with pytest.raises(
        ValueError, match="Invalid index_table: " "missing address column"
    ):
        weave.load_mongo(
            pd.DataFrame({"uuid": ["1234"], "basket_type": ["type"]}),
            file_system=set_up.file_system,
            collection=set_up.test_collection,
        )


@pytest.mark.skipif(
    "pymongo" not in sys.modules or not os.environ.get("MONGODB_HOST", False),
    reason="Pymongo required for this test",
)
def test_load_mongo_check_dataframe_for_basket_type(set_up):
    """Test that load_mongo prevents loading data with missing basket type."""
    with pytest.raises(
        ValueError, match="Invalid index_table: " "missing basket_type column"
    ):
        weave.load_mongo(
            pd.DataFrame({"uuid": ["1234"], "address": ["path"]}),
            file_system=set_up.file_system,
            collection=set_up.test_collection,
        )


@pytest.mark.skipif(
    "pymongo" not in sys.modules or not os.environ.get("MONGODB_HOST", False),
    reason="Pymongo required for this test",
)
def test_load_mongo_check_for_duplicate_uuid(set_up):
    """Test duplicate metadata won't be uploaded to mongoDB, based on the UUID.
    """

    test_uuid = "1234"

    # Load metadata twice and ensure there's only one instance
    index_table = weave.index.create_index.create_index_from_fs(
        set_up.pantry_path, set_up.file_system
    )
    weave.load_mongo(
        index_table,
        file_system=set_up.file_system,
        collection=set_up.test_collection,
    )
    weave.load_mongo(
        index_table,
        file_system=set_up.file_system,
        collection=set_up.test_collection,
    )
    count = set_up.mongodb[set_up.test_collection].count_documents(
        {"uuid": test_uuid}
    )
    assert count == 1, "duplicate uuid inserted"



"""This script contains tests that test that pytest is deleting baskets
correctly."""

import os
import sys

from fsspec.implementations.local import LocalFileSystem
# Try-Except required to make pyodbc an optional dependency.
# Ignore pylint. This is used to explicitly show the optional dependency.
# pylint: disable=duplicate-code
try:
    import pyodbc
except ImportError:
    _HAS_PYODBC = False
else:
    _HAS_PYODBC = True
import pytest
import s3fs

from weave.tests.pytest_resources import PantryForTest


s3fs = s3fs.S3FileSystem(
    client_kwargs={"endpoint_url": os.environ["S3_ENDPOINT"]}
)
local_fs = LocalFileSystem()


# Test with two different fsspec file systems (above).
@pytest.fixture(
    name="set_up_tb_no_cleanup",
    params=[s3fs, local_fs],
    ids=["S3FileSystem", "LocalFileSystem"],
)
def fixture_set_up_tb_no_cleanup(request, tmpdir):
    """Sets up test basket fixture."""
    file_system = request.param
    temp_basket = PantryForTest(tmpdir, file_system)
    # Purposefully don't clean up pantry, it will be cleaned up in the test.
    return temp_basket


def test_weave_pytest_suffix(set_up_tb_no_cleanup):
    """Test that env var suffix works, and pantrys are still deleted."""
    # Check pantry name includes suffix if applicable.
    suffix = os.environ.get("WEAVE_PYTEST_SUFFIX", "")
    assert set_up_tb_no_cleanup.pantry_path == f"pytest-temp-pantry{suffix}"

    # Check the pantry was made.
    assert set_up_tb_no_cleanup.file_system.exists(
        set_up_tb_no_cleanup.pantry_path
    )

    # Cleanup the pantry.
    set_up_tb_no_cleanup.cleanup_pantry()

    # Check the pantry is actually deleted.
    assert not set_up_tb_no_cleanup.file_system.exists(
        set_up_tb_no_cleanup.pantry_path
    )


# Skip tests if pyodbc is not installed.
@pytest.mark.skipif(
    "pyodbc" not in sys.modules or not _HAS_PYODBC
    or not os.environ["MSSQL_PASSWORD"] or not os.environ["MSSQL_HOST"],
    reason="Module 'pyodbc' required for this test "
    "AND env variables: 'MSSQL_HOST', 'MSSQL_PASSWORD'",
)
def test_github_cicd_sql_server():
    """Test that the MS SQL Server is properly setup in CICD."""
    # Documentation for mssql container:
    # https://hub.docker.com/_/microsoft-mssql-server
    # Default System Administrator username is "sa"
    username = "sa"
    mssql_password = os.environ["MSSQL_PASSWORD"]
    server = os.environ["MSSQL_HOST"]
    database = "tempdb"

    # Pylint has a problem recognizing 'connect' as a valid member function
    # so we ignore that here.
    # pylint: disable-next=c-extension-no-member
    con = pyodbc.connect(
        "DRIVER={ODBC Driver 18 for SQL Server};"
        f"SERVER={server};"
        f"DATABASE={database};"
        f"UID={username};"
        f"PWD={mssql_password};"
        f"Encrypt=no;"
    )
    cur = con.cursor()

    # Create a temporary table for testing.
    cur.execute("""
    IF NOT EXISTS (
        SELECT * FROM sys.tables t
        JOIN sys.schemas s ON (t.schema_id = s.schema_id)
        WHERE s.name = 'dbo' AND t.name = 'test_table'
    )
    CREATE TABLE dbo.test_table (
        uuid varchar(64),
        num int
    );
    """)

    # Insert a test value, and then check we can retrieve the value.
    cur.execute("""
        INSERT INTO dbo.test_table (uuid, num) VALUES ('0001', 1);
    """)
    cur.commit()
    assert cur.execute("SELECT * FROM dbo.test_table").fetchall() != []

    # Delete the test value, and then check it was actually deleted.
    cur.execute("""DELETE FROM dbo.test_table WHERE uuid = '0001';""")
    cur.commit()
    assert cur.execute("SELECT * FROM dbo.test_table;").fetchall() == []

    cur.execute("""DROP TABLE dbo.test_table;""")
    cur.commit()



"""Contains functions and classes used by uploader.py's upload function."""

import json
import os
import warnings

import jsonschema
from jsonschema import validate

from .config import manifest_schema, supplement_schema, prohibited_filenames


def validate_pantry(pantry):
    """Starts the validation process off based off the name of the pantry.

    Validates that the pantry actually exists at the location given.
    If there is a pantry that exists, check it and every subdirectory by
    calling check_level.

    Parameters
    ----------
    pantry: weave.Pantry
        Pantry to be validated.

    Returns
    ----------
    A list of all invalid basket locations (will return an empty list if
    no warnings are raised).
    """

    if not pantry.file_system.exists(pantry.pantry_path):
        raise ValueError(
            f"Invalid pantry Path. Pantry does not exist at: "
            f"{pantry.pantry_path}"
        )

    # Catching the warnings that are shown from calling
    # generate_index() to prevent showing the same warning twice
    with warnings.catch_warnings(record=True):
        try:
            pantry.index.generate_index()
        except json.decoder.JSONDecodeError as error:
            raise ValueError(
                f"Pantry could not be loaded into index: {error}"
            ) from error

    # Call check level using the pantry_name as the path
    with warnings.catch_warnings(record=True) as warn:
        _check_level(pantry.pantry_path, pantry=pantry)
        # Iterate through warn and return the list of warning messages.
        # Enumerate does not work here. prefer to use range and len
        # pylint: disable-next=consider-using-enumerate
        warning_list = [warn[i].message for i in range(len(warn))]
        return warning_list


def _check_level(current_dir, **kwargs):
    """Check all immediate subdirs in dir, check for manifest.

    Checks all the immediate subdirectories and files in the given directory
    to see if there is a basket_manifest.json file. If there is a manifest,
    it's a basket and must be validated.
    If there is a directory found, then recursively call check_level with that
    found directory to find if there is a basket in any directory.

    Parameters
    ----------
    current_dir: string
        the current directory that we want to search all files and
        directories of
    **pantry: weave.Pantry (required)
        Pantry object representing the pantry to validate.
    **in_basket: bool (optional)
        This is a flag to signify that we are in a basket
        and we are looking for a nested basket now.

    Returns
    ----------
    bool that comes from:
        a validate_basket() if there is a basket found
        a check_level() if there is a directory found
        a true if a manifest is found while inside another basket
        a default true if no basket is found
    """

    # Collect kwargs
    pantry = kwargs.get("pantry")
    file_system = pantry.file_system
    in_basket = kwargs.get("in_basket", False)

    if not file_system.exists(current_dir):
        raise ValueError(
            f"Invalid Path. No file or directory found at: {current_dir}"
        )


    manifest_path = os.path.join(current_dir, "basket_manifest.json")

    # If a manifest exists, it's a basket, validate it
    if file_system.exists(manifest_path):
        # If there is another manifest inside a basket,
        # the nested basket does not need to be validated
        if in_basket:
            return True
        return _validate_basket(current_dir, pantry)

    # Go through all the other files, if it's a directory, check it
    dirs_and_files = file_system.ls(path=current_dir, refresh=True)

    for file_or_dir in dirs_and_files:
        file_type = file_system.info(file_or_dir)["type"]

        if file_type == "directory":
            # If directory is a basket, check everything under it
            # for a manifest and return True
            # This will return True to the _validate_basket
            # and throw an error or warning
            if in_basket:
                return _check_level(file_or_dir,
                                    pantry=pantry,
                                    in_basket=in_basket)
            # If directory is not a basket, check all files in the
            # current dir. If everything is valid, _check_level returns True
            # If it isn't valid, return False
            # and continue looking at all the sub-directories
            if not _check_level(file_or_dir,
                                pantry=pantry,
                                in_basket=in_basket):
                return False

    # This is the default backup return
    # Return True if directory is not a basket because
    # it is valid to have no baskets
    # Return False if directory is a basket because
    # check_level did not find another basket
    return not in_basket


def _validate_basket(basket_dir, pantry):
    """Takes the root directory of a basket and validates it.

    Validation means there is a required basket_manifest.json and
    basket_supplment.json, and an optional basket_metadata.json.
    The manifest and supplement are required to follow a certain schema,
    which is defined in config.py.
    All three: manifest, supplement, and metadata are also verified by being
    able to be read into a python dictionary.

    If there are any directories found inside this basket, run check_basket()
    on them to see if there is another basket inside this basket. If there is
    another basket, raise an error or warning for invalid pantry.

    If the basket is ever invalid, raise an error or warning.
    If the basket is valid return True.

    Parameters
    ----------
    basket_dir: str
        The path in the file system to the basket root directory.
    pantry: weave.Pantry
        Pantry object representing the pantry to validate.

    Returns
    ----------
    Boolean that is True when the basket is valid.
    If the Basket is invalid, raise an error or warning.
    """

    manifest_path = os.path.join(basket_dir, "basket_manifest.json")
    supplement_path = os.path.join(basket_dir, "basket_supplement.json")

    # A valid basket has both manifest and supplement
    # If for some reason the manifest is gone,
    # either the directory is wrong,
    # or this function is incorrectly called,
    if not pantry.file_system.exists(manifest_path):
        raise FileNotFoundError(f"Invalid Path. "
                                f"No Basket found at: {basket_dir}")

    if not pantry.file_system.exists(supplement_path):
        warnings.warn(UserWarning(
            "Invalid Basket. No Supplement file found at: ", basket_dir))

    files_in_basket = pantry.file_system.ls(path=basket_dir, refresh=True)

    basenames = [os.path.basename(x) for x in files_in_basket]

    if set(prohibited_filenames) == set(basenames):
        # Basket with no files, check if it's a metadata-only basket
        _check_metadata_only(files_in_basket, pantry)

    for file in files_in_basket:
        _, file_name = os.path.split(file)
        {
            "basket_manifest.json": _handle_manifest,
            "basket_supplement.json": _handle_supplement,
            "basket_metadata.json": _handle_metadata
        }.get(
            file_name, _handle_none_of_the_above
        )(file, pantry)

    # Default return True if there are no problems with this basket
    return True


def _handle_manifest(file, pantry):
    """Handles case if manifest.

    Parameters:
    -----------
    file: str
        Path to the file.
    pantry: weave.Pantry
        Pantry object representing the pantry to validate.
    """

    try:
        # Make sure it can be loaded, valid schema, and valid parent_uuids
        data = json.load(pantry.file_system.open(file))
        validate(instance=data, schema=manifest_schema)
        _validate_parent_uuids(data, pantry)

    except jsonschema.exceptions.ValidationError:
        warnings.warn(UserWarning(
            "Invalid Basket. "
            "Manifest Schema does not match at: ", file
        ))

    except json.decoder.JSONDecodeError:
        warnings.warn(UserWarning(
            "Invalid Basket. "
            "Manifest could not be loaded into json at: ", file
        ))


def _handle_supplement(file, pantry):
    """Handles case if supplement.

    Parameters:
    -----------
    file: str
        Path to the file.
    pantry: weave.Pantry
        Pantry object representing the pantry to validate.
    """

    try:
        # These two lines make sure it can be read and is valid schema
        data = json.load(pantry.file_system.open(file))
        validate(instance=data, schema=supplement_schema)
        basket_dir, _ = os.path.split(file)
        _validate_supplement_files(basket_dir, data, pantry)

    except jsonschema.exceptions.ValidationError:
        warnings.warn(UserWarning(
            "Invalid Basket. "
            "Supplement Schema does not match at: ", file
        ))

    except json.decoder.JSONDecodeError:
        warnings.warn(UserWarning(
            "Invalid Basket. "
            "Supplement could not be loaded into json at: ", file
        ))


def _handle_metadata(file, pantry):
    """Handles case if metadata.

    Parameters:
    -----------
    file: str
        Path to the file.
    pantry: weave.Pantry
        Pantry object representing the pantry to validate.
    """

    try:
        json.load(pantry.file_system.open(file))

    except json.decoder.JSONDecodeError:
        warnings.warn(UserWarning(
            "Invalid Basket. "
            "Metadata could not be loaded into json at: ", file
        ))


def _handle_none_of_the_above(file, pantry):
    """Handles case if none of the above.

    Parameters:
    -----------
    file: str
        Path to the file.
    pantry: weave.Pantry
        Pantry object representing the pantry to validate.
    """

    basket_dir, _ = os.path.split(file)

    if pantry.file_system.info(file)['type'] == 'directory':
        if _check_level(file,
                        pantry=pantry,
                        in_basket=True):
            warnings.warn(UserWarning(
                "Invalid Basket. Manifest File "
                "found in sub directory of basket at: ", basket_dir
            ))


def _check_metadata_only(files_in_basket, pantry):
    """Checks if a basket is a metadata-only basket.

    Parameters:
    -----------
    files_in_basket: [str]
        List of all file paths in a basket.
    pantry: weave.Pantry
        Pantry object representing the pantry to validate.
    """
    for file in files_in_basket:
        if file.endswith("basket_manifest.json"):
            man_data = json.load(pantry.file_system.open(file))

        if file.endswith("basket_supplement.json"):
            supp_data = json.load(pantry.file_system.open(file))

        if file.endswith("basket_metadata.json"):
            meta_data = json.load(pantry.file_system.open(file))

    # Check that it is a metadata-only basket by checking 3 things:
    # 1. metadata is not empty
    # 2. Basket has parent_uuids
    # 3. No files were uploaded (this is checked twice)
    if not (meta_data and
            not supp_data["integrity_data"] and
            man_data["parent_uuids"]):
        # Raise a warning that there are no files uploaded, but it
        # is not considered to be a metadata-only basket.
        warnings.warn(UserWarning(
                "Invalid Basket. No files in basket and criteria not met for "
                "metadata-only basket. ", man_data["uuid"]
            ))


def _validate_parent_uuids(data, pantry):
    """Validate that all the parent_uuids from the manifest exist in the
    pantry.

    If there are parent uuids that don't actually exist in the pantry,
    a warning will be raised for each of those, along with the
    basket's uuid where the error occurred.

    Parameters
    ----------
    data: dict
        The dictionary that contains the data of the manifest.json.
    pantry: weave.Pantry
        Pantry object representing the pantry to validate.
    """

    # If there are no parent uuids in the manifest, no need to check anything
    if len(data["parent_uuids"]) == 0:
        return

    found_parents = list(pantry.index.get_rows(data['parent_uuids'])['uuid'])
    missing_uids = (
        [uuid for uuid in data["parent_uuids"] if uuid not in found_parents]
    )

    if missing_uids:
        warnings.warn(f"The uuids: {missing_uids} were not found in the "
                      f"index, which was found inside basket: {data['uuid']}")


def _validate_supplement_files(basket_dir, data, pantry):
    """Validate the files listed in the supplement's integrity_data.

    Parameters
    ----------
    basket_dir: str
        The path to the current working basket.
    data: dictionary
        The dictionary that contains the data of the supplement.json.
    pantry: weave.Pantry
        The pantry to validate.
    """
    pantry_path = pantry.pantry_path
    sys_file_list = pantry.file_system.find(path=basket_dir, withdirs=False)

    manifest_path = os.path.join(basket_dir, "basket_manifest.json")
    supplement_path = os.path.join(basket_dir, "basket_supplement.json")
    metadata_path = os.path.join(basket_dir, "basket_metadata.json")

    # Grab all the files, but remove manifest, supplement, and metadata
    system_file_list = [
        file for file in sys_file_list if file not in [manifest_path,
                                                       supplement_path,
                                                       metadata_path]
    ]

    supp_file_list = [file["upload_path"] for file in data["integrity_data"]]

    # Remove path up until the pantry directory in both lists
    system_file_list = [
        file[file.find(pantry_path):] for file in system_file_list
    ]
    supp_file_list = [file[file.find(pantry_path):] for file in supp_file_list]

    system_file_set = set(system_file_list)
    supp_file_set = set(supp_file_list)

    files_not_in_system = supp_file_set - system_file_set
    files_not_in_supp = system_file_set - supp_file_set

    for file in files_not_in_system:
        warnings.warn(
            UserWarning("File listed in the basket_supplement.json does not "
                        "exist in the file system: ", file)
        )

    for file in files_not_in_supp:
        warnings.warn(
            UserWarning("File found in the file system is not listed in "
                        "the basket_supplement.json: ", file)
        )



"""This script contains tests that test that pytest is deleting baskets
correctly."""

import os
from unittest import mock

# Try-Except required to make sqlalchemy an optional dependency.
try:
    # For the sake of explicitly showing that sqlalchemy is optional, import
    # it here, even though it is not currently used in this file.
    # Pylint ignore the next unused-import pylint warning.
    # Also inline ruff ignore unused import (F401)
    # pylint: disable=unused-import
    # pyodbc is imported here because sqlalchemy requires it.
    import pyodbc # noqa: F401
    import sqlalchemy as sqla # noqa: F401
    # pylint: enable=unused-import
except ImportError:
    _HAS_REQUIRED_DEPS = False
else:
    _HAS_REQUIRED_DEPS = True
import pytest
import s3fs
from fsspec.implementations.local import LocalFileSystem

from weave.index.index_sql import IndexSQL
from weave.tests.pytest_resources import IndexForTest
from weave.tests.pytest_resources import PantryForTest
from weave.tests.pytest_resources import get_sample_basket_df

s3fs = s3fs.S3FileSystem(
    client_kwargs={"endpoint_url": os.environ["S3_ENDPOINT"]}
)
local_fs = LocalFileSystem()


# Test with two different fsspec file systems (above).
@pytest.fixture(
    name="test_pantry",
    params=[s3fs, local_fs],
    ids=["S3FileSystem", "LocalFileSystem"],
)
def fixture_test_pantry(request, tmpdir):
    """Sets up test pantry for the tests"""
    file_system = request.param
    test_pantry = PantryForTest(tmpdir, file_system)
    yield test_pantry
    test_pantry.cleanup_pantry()


@pytest.fixture(
    name="test_index",
    params=[IndexSQL],
    ids=["IndexSQL"],
)
def fixture_test_index(request):
    """Sets up test index for the tests"""
    index_constructor = request.param
    test_index = IndexForTest(index_constructor, local_fs)
    yield test_index
    test_index.cleanup_index()


# Skip tests if sqlalchemy is not installed.
@pytest.mark.skipif(
    not _HAS_REQUIRED_DEPS
    or not os.environ.get("MSSQL_PASSWORD", False),
    reason="Modules: 'pyodbc', 'sqlalchemy' required for this test "
    "AND env variables: 'MSSQL_HOST', 'MSSQL_PASSWORD'",
)
# Mock the environment variables for the test.
@mock.patch.dict(os.environ, {}, clear=True)
def test_index_sql_no_env_vars():
    """Test that the SQL Index will raise an error if the required env vars
    are not set.
    """
    with pytest.raises(KeyError) as err:
        IndexSQL(LocalFileSystem(), "weave-test-pantry")

    assert (
        str(err.value) == "'The following environment variables must be set to"
        " use this class: MSSQL_HOST, MSSQL_USERNAME, MSSQL_PASSWORD.'"
    )


# Skip tests if sqlalchemy is not installed.
@pytest.mark.skipif(
    not _HAS_REQUIRED_DEPS
    or not os.environ.get("MSSQL_PASSWORD", False),
    reason="Modules: 'pyodbc', 'sqlalchemy' required for this test "
    "AND env variables: 'MSSQL_HOST', 'MSSQL_PASSWORD'",
)
def test_index_sql_properties_are_read_only():
    """Test that the properties of the SQL Index (database_name, pantry_schema)
    are read only, and cannot be changed during execution.
    """
    pantry_path = (
        "pytest-temp-pantry"
        f"{os.environ.get('WEAVE_PYTEST_SUFFIX', '')}"
    )

    ind = IndexSQL(LocalFileSystem(), pantry_path)

    original_db_name = "weave_db"
    original_schema_name = pantry_path.replace("-", "_")

    with pytest.raises(AttributeError):
        ind.database_name = "new_db_name"

    with pytest.raises(AttributeError):
        ind.pantry_schema = "new_schema_name"

    assert ind.database_name == original_db_name
    assert ind.pantry_schema == original_schema_name


# Skip tests if sqlalchemy is not installed.
@pytest.mark.skipif(
    not _HAS_REQUIRED_DEPS
    or not os.environ.get("MSSQL_PASSWORD", False),
    reason="Modules: 'pyodbc', 'sqlalchemy' required for this test "
    "AND env variables: 'MSSQL_HOST', 'MSSQL_PASSWORD'",
)
def test_index_sql_tracks_different_pantries():
    """Test that the SQL Index will track different baskets using schemas."""
    sample_basket_df = get_sample_basket_df()
    uuid = str(sample_basket_df["uuid"].iloc[0])

    pantry_path = (
        "pytest-temp-pantry"
        f"{os.environ.get('WEAVE_PYTEST_SUFFIX', '')}"
    )

    ind_1 = IndexSQL(LocalFileSystem(), pantry_path + "_1")
    ind_2 = IndexSQL(LocalFileSystem(), pantry_path + "_2")

    # Perform tracks and untracks on both indices, and ensure they are correct.
    ind_1.track_basket(sample_basket_df)
    assert len(ind_1) == 1
    assert len(ind_2) == 0

    ind_2.track_basket(sample_basket_df)
    assert len(ind_1) == 1
    assert len(ind_2) == 1

    ind_1.untrack_basket(uuid)
    assert len(ind_1) == 0
    assert len(ind_2) == 1

    ind_2.untrack_basket(uuid)
    assert len(ind_1) == 0
    assert len(ind_2) == 0

    # Need to manually clean up the database in this test.
    ind_1.execute_sql(f"DROP TABLE {ind_1.pantry_schema}.pantry_index;",
        commit=True)
    ind_1.execute_sql(f"DROP TABLE {ind_1.pantry_schema}.parent_uuids;",
        commit=True)
    ind_1.execute_sql(f"DROP SCHEMA {ind_1.pantry_schema};",
        commit=True)
    ind_2.execute_sql(f"DROP TABLE {ind_2.pantry_schema}.pantry_index;",
        commit=True)
    ind_2.execute_sql(f"DROP TABLE {ind_2.pantry_schema}.parent_uuids;",
        commit=True)
    ind_2.execute_sql(f"DROP SCHEMA {ind_2.pantry_schema};",
        commit=True)


# Skip tests if sqlalchemy is not installed.
@pytest.mark.skipif(
    not _HAS_REQUIRED_DEPS
    or not os.environ.get("MSSQL_PASSWORD", False),
    reason="Modules: 'pyodbc', 'sqlalchemy' required for this test "
    "AND env variables: 'MSSQL_HOST', 'MSSQL_PASSWORD'",
)
def test_index_sql_track_basket_adds_to_parent_uuids(test_index):
    """Test that track_basket adds necessary rows to the parent_uuids table."""
    sample_basket_df = get_sample_basket_df()
    uuid = "1000"
    sample_basket_df["uuid"] = uuid

    # Add uuids to the parent_uuids of the df.
    sample_basket_df["parent_uuids"] = [["0001", "0002", "0003"]]

    # Track the basket.
    test_index.index.track_basket(sample_basket_df)

    ind = test_index.index
    # Manually check the parent_uuids table.
    rows, _ = ind.execute_sql(
        f"SELECT * FROM {test_index.index.pantry_schema}.parent_uuids"
    )

    print(rows)

    # Check we have the expected values.
    assert len(rows) == 3
    assert str(rows[0]) == str(("1000", "0001"))
    assert str(rows[1]) == str(("1000", "0002"))
    assert str(rows[2]) == str(("1000", "0003"))

    # Untrack the basket and ensure values are removed from parent_uuids table.
    test_index.index.untrack_basket(uuid)
    rows, _ = ind.execute_sql(
        f"SELECT * FROM {test_index.index.pantry_schema}.parent_uuids"
    )
    assert len(rows) == 0



"""Provides configuration settings used by Weave."""

import os

import s3fs
# Try-Except required to make pymongo an optional dependency.
try:
    import pymongo
except ImportError:
    _HAS_PYMONGO = False
else:
    _HAS_PYMONGO = True


# Filenames not allowed to be added to the basket.
# These files are taken for specific weave purposes.
prohibited_filenames = [
    "basket_manifest.json",
    "basket_metadata.json",
    "basket_supplement.json",
]

# basket_manifest must follow this schema
manifest_schema = {
    "properties": {
        "uuid": {"type": "string"},
        "upload_time": {"type": "string"},
        "parent_uuids": {"type": "array", "items": {"type": "string"}},
        "basket_type": {"type": "string"},
        "label": {"type": "string"},
        "weave_version": {"type": "string"}
    },
    "required": [
        "uuid",
        "upload_time",
        "parent_uuids",
        "basket_type",
        "label",
    ],
    "additionalProperties": False,
}

# basket_supplement must follow this schema
supplement_schema = {
    "properties": {
        "upload_items": {
            "type": "array",
            "minItems": 0,
            "items": {
                "type": "object",
                "properties": {
                    "path": {"type": "string"},
                    "stub": {"type": "boolean"},
                },
                "required": ["path", "stub"],
                "additionalProperties": False,
            },
        },
        "integrity_data": {
            "type": "array",
            "minItems": 0,
            "items": {
                "type": "object",
                "properties": {
                    "file_size": {"type": "number"},
                    "hash": {"type": "string"},
                    "access_date": {"type": "string"},
                    "source_path": {"type": "string"},
                    "byte_count": {"type": "number"},
                    "stub": {"type": "boolean"},
                    "upload_path": {"type": "string"},
                },
                "required": [
                    "file_size",
                    "hash",
                    "access_date",
                    "source_path",
                    "byte_count",
                    "stub",
                    "upload_path",
                ],
                "additionalProperties": False,
            },
            "required": ["type"],
            "additionalProperties": False,
        },
    },
    "required": ["upload_items", "integrity_data"],
    "additionalProperties": False,
}


def get_index_column_names():
    """Return index column names."""
    return ["uuid", "upload_time", "parent_uuids", "basket_type", "label",
            "weave_version", "address", "storage_type"]


def get_file_system():
    """Get the filesystem to be used for storing baskets."""
    return s3fs.S3FileSystem(
        client_kwargs={"endpoint_url": os.environ["S3_ENDPOINT"]}
    )


def get_mongo_db():
    """Get the mongodb client to be used for metadata search."""

    if not _HAS_PYMONGO:
        raise ImportError("Missing Dependency. The package 'pymongo' "
                          "is required to use this function")

    # If MONGODB_HOST, USERNAME and PASSWORD are provided as environment
    # variables, initialize the mongo client with the provided
    # credentials. Else defer to default credentials for OPAL.
    if (
        "MONGODB_HOST" in os.environ
        and "MONGODB_USERNAME" in os.environ
        and "MONGODB_PASSWORD" in os.environ
    ):
        client = pymongo.MongoClient(
            host=os.environ["MONGODB_HOST"],
            username=os.environ["MONGODB_USERNAME"],
            password=os.environ["MONGODB_PASSWORD"],
        )
    else:
        client = pymongo.MongoClient(
            "mongodb", username="root", password="example"
        )
    return client
